"""
    Implementation of the `unsigned-varint spec <https://github.com/multiformats/unsigned-varint>`_.

    Suggested usage:

    >>> from multiformats import varint
"""

from io import BufferedIOBase
from typing import BinaryIO, cast, List, Optional, overload, Tuple, Union, TypeVar
from typing_extensions import Final
from typing_validation import validate

_max_num_bytes: int = 9

BytesLike = Union[bytes, bytearray, memoryview]
""" Type alias for bytes-like objects. """

byteslike: Final = (bytes, bytearray, memoryview)
""" Tuple of bytes-like objects types (for use with :obj:`isinstance` checks). """

def encode(x: int) -> bytes:
    """
        Encodes a non-negative integer as an unsigned varint, returning the encoded bytes.

        Example usage:

        >>> from multiformats import varint
        >>> varint.encode(128)
        b'\\x80\\x01'

        :param x: the non-negative integer to encode
        :type x: :obj:`int`

        :raises ValueError: if `x < 0` (varints encode unsigned integers)
        :raises ValueError: if `x >= 2**63` (from specs, varints are limited to 9 bytes)
    """
    validate(x, int)
    if x < 0:
        raise ValueError("Integer is negative.")
    varint_bytelist: List[int] = []
    while True:
        next_byte = x & 0b0111_1111
        x >>= 7
        if x > 0:
            varint_bytelist.append(next_byte | 0b1000_0000)
        else:
            varint_bytelist.append(next_byte)
            break
        if len(varint_bytelist) >= _max_num_bytes:
            raise ValueError(f"Varints must be at most {_max_num_bytes} bytes long.")
    return bytes(varint_bytelist)


def decode(b: Union[BytesLike, BufferedIOBase, BinaryIO]) -> int:
    """
        Decodes an unsigned varint from a bytes-like object or a buffered binary stream.

        - if a stream is passed, only the bytes encoding the varint are read from it
        - if a `bytes`-like object is passed, the varint encoding must use all bytes

        Example usage with bytes:

        >>> from multiformats import varint
        >>> varint.decode(b'\\x80\\x01')
        128

        Example usage with streams, for the (typical) situation where the varint is only part of the data:

        >>> from io import BytesIO
        >>> stream = BytesIO(b"\\x80\\x01\\x12\\xff\\x01")
        >>> varint.decode(stream)
        128
        >>> stream.read() # what's left in the stream
        b'\\x12\\xff\\x01'

        :param b: the bytes-like object or stream from which to decode a varint
        :type b: :obj:`~multiformats.varint.BytesLike`, :obj:`~io.BufferedIOBase` or :obj:`~typing.BinaryIO`

        :raises ValueError: if the input contains no bytes (from specs, the number 0 is encoded as ``0b00000000``)
        :raises ValueError: if the 9th byte of the input is a continuation byte (from specs, no number >= 2**63 is allowed)
        :raises ValueError: if the last byte of the input is a continuation byte (invalid format)
        :raises ValueError: if the decoded integer could be encoded in fewer bytes than were read (from specs, encoding must be minimal)
        :raises ValueError: if the input is a bytes-like object and the number of bytes used by the encoding is fewer than its length

        The last point is a designed choice aimed to reduce errors when decoding fixed-length bytestrings (rather than streams).
        If this behaviour is undesirable, consider using `decode_head` instead.

    """
    x, num_bytes_read, _ = decode_raw(b)
    if isinstance(b, byteslike) and len(b) > num_bytes_read:
        raise ValueError("A bytes-like object was passed, but not all bytes were used by the encoding.")
    return x

def _no_next_byte_error(num_bytes_read: int) -> ValueError:
    if num_bytes_read == 0:
        return ValueError("Varints must be at least 1 byte long.")
    return ValueError(f"Byte #{num_bytes_read-1} was a continuation byte, but byte #{num_bytes_read} not available.")

_BufferedIOT = TypeVar("_BufferedIOT", bound=BufferedIOBase)
_BinaryIOT = TypeVar("_BinaryIOT", bound=BinaryIO)

@overload
def decode_raw(b: BytesLike) -> Tuple[int, int, memoryview]:
    ...

@overload
def decode_raw(b: _BufferedIOT) -> Tuple[int, int, _BufferedIOT]:
    ...

@overload
def decode_raw(b: _BinaryIOT) -> Tuple[int, int, _BinaryIOT]:
    ...

def decode_raw(b: Union[BytesLike, BufferedIOBase, BinaryIO]) -> Tuple[int, int, Union[memoryview, BufferedIOBase, BinaryIO]]:
    """
        Specialised version of :func:`~multiformats.varint.decode` for partial decoding, returning a pair ``(x, n)`` of
        the decoded varint ``x`` and the number ``n`` of bytes read from the start and/or consumed from the stream.
        Unlike :func:`~multiformats.varint.decode`, this function doesn't raise `ValueError` in case not all bytes are read in the process.

        Example usage with bytes:

        >>> bs = b"\\x80\\x01\\x12\\xff\\x01"
        >>> x, n, m = varint.decode_raw(bs)
        >>> x
        128
        >>> n
        2
        # read first 2 bytes: b"\\x80\\x01"
        >>> m
        <memory at 0x000001A6E55DDA00>
        >>> bytes(m)
        b'\\x12\\xff\\x01'
        # memoryview on remaining bytes
        # note: bytes(m) did not consume the bytes

        Example usage with streams, for the (typical) situation where the varint is only part of the data:

        >>> from io import BytesIO
        >>> stream = BytesIO(b"\\x80\\x01\\x12\\xff\\x01")
        >>> x, n = varint.decode_head(stream)
        >>> x
        128
        >>> n
        2
        # read first 2 bytes: b"\\x80\\x01"
        >>> m
        <_io.BytesIO object at 0x000001A6E554BBD0>
        >>> m == stream
        True
        # original stream returned, containing remaining bytes
        >>> stream.read()
        b'\\x12\\xff\\x01'
        # 2 bytes were consumed decoding the varint, so 3 bytes were left in the stream
        # note: stream.read() consumed the bytes

        :param b: the bytes-like object or stream from which to decode a varint
        :type b: :obj:`~multiformats.varint.BytesLike`, :obj:`~io.BufferedIOBase` or :obj:`~typing.BinaryIO`

        :raises ValueError: same reasons as :func:`~multiformats.varint.decode`, except for the last (where no error is raised)

    """
    stream_mode: Optional[type]
    if isinstance(b, BufferedIOBase):
        stream_mode = BufferedIOBase
        validate(b, BufferedIOBase)
    elif isinstance(b, BinaryIO):
        stream_mode = BinaryIO
        validate(b, BinaryIO)
    else:
        stream_mode = None
        validate(b, BytesLike)
    expect_next = True
    num_bytes_read = 0
    x = 0
    while expect_next:
        if stream_mode is not None:
            _next_byte: bytes = cast(Union[BufferedIOBase, BinaryIO], b).read(1)
            if len(_next_byte) == 0:
                raise _no_next_byte_error(num_bytes_read)
            next_byte: int = _next_byte[0]
        else:
            if num_bytes_read >= len(cast(BytesLike, b)):
                raise _no_next_byte_error(num_bytes_read)
            next_byte = cast(BytesLike, b)[num_bytes_read]
        x += (next_byte & 0b0111_1111) << (7 * num_bytes_read)
        expect_next = (next_byte >> 7) == 0b1
        num_bytes_read += 1
        if expect_next and num_bytes_read >= _max_num_bytes:
            raise ValueError(f"Varints must be at most {_max_num_bytes} bytes long.")
    if num_bytes_read > 1 and x < 2**(7*(num_bytes_read-1)):
        raise ValueError(f"Number {x} was not minimally encoded (as a {num_bytes_read} bytes varint).")
    if stream_mode is not None:
        return x, num_bytes_read, cast(Union[BufferedIOBase, BinaryIO], b)
    return x, num_bytes_read, memoryview(cast(BytesLike, b))[num_bytes_read:]



"""  Tests for the `multiformats.multicodec` module. """

from typing import Dict, List, Union

import pytest

import multiformats_config
from multiformats import multicodec
from multiformats.multicodec import Multicodec

exists_cases = [
    "identity",
    0x00
]

@pytest.mark.parametrize("name_or_code", exists_cases)
def test_exists(name_or_code: Union[str, int]) -> None:
    """ Tests `multicodec.exists`. """
    if isinstance(name_or_code, str):
        assert multicodec.exists(name_or_code)
    else:
        assert multicodec.exists(code=name_or_code)

get_cases = {
    "identity": {
        "name": "identity",
        "code": 0x00,
        "hexcode": "0x00",
        "tag": "multihash",
        "status": "permanent",
        "description": "raw binary",
    }
}

@pytest.mark.parametrize("name, info", get_cases.items())
def test_get(name: str, info: Dict[str, Union[str, int]]) -> None:
    """ Tests `multicodec.get`. """
    m = multicodec.get(name)
    for k, v in info.items():
        assert hasattr(m, k)
        assert getattr(m, k) == v

json_cases = [
    {
        "name": "my-codec",
        "tag": "private",
        "code": "0x300001",
        "status": "draft",
        "description": "my private codec"
    }
]

@pytest.mark.parametrize("m_json", json_cases)
def test_multicodec_contructor(m_json: Dict[str, str]) -> None:
    """ Tests `Multicodec.from_json`. """
    m = Multicodec(**m_json)
    assert m.name == m_json["name"]
    assert m.code == int(m_json["code"], base=16)
    assert m.hexcode == m_json["code"]
    assert m.status == m_json["status"]
    assert m.description == m_json["description"]
    assert str(m) == f"Multicodec({', '.join(f'{k}={repr(v)}' for k, v in m_json.items())})"

@pytest.mark.parametrize("m_json", json_cases)
def test_multicodec_to_json(m_json: Dict[str, str]) -> None:
    """ Tests `Multicodec.to_json`. """
    assert Multicodec(**m_json).to_json() == m_json

@pytest.mark.parametrize("m_json", json_cases)
def test_register(m_json: Dict[str, str]) -> None:
    """ Tests `multicodec.register`. """
    m = Multicodec(**m_json)
    assert not multicodec.exists(m.name)
    assert not multicodec.exists(code=m.code)
    multicodec.register(m)
    assert multicodec.exists(m.name)
    assert multicodec.exists(code=m.code)
    assert multicodec.get(m.name) == multicodec.get(code=m.code) == m

@pytest.mark.parametrize("m_json", json_cases)
def test_unregister(m_json: Dict[str, str]) -> None:
    """ Tests `multicodec.get`. """
    m = Multicodec(**m_json)
    assert multicodec.exists(m.name)
    assert multicodec.exists(code=m.code)
    m2_json = {
        **m_json,
        "name": m_json["name"]+"-2",
        "code": hex(int(m_json["code"], base=16)+1)
    }
    m2 = Multicodec(**m2_json)
    assert not multicodec.exists(m2.name)
    assert not multicodec.exists(code=m2.code)
    multicodec.register(m2)
    multicodec.unregister(m.name)
    assert not multicodec.exists(m.name)
    assert not multicodec.exists(code=m.code)
    multicodec.unregister(code=m2.code)
    assert not multicodec.exists(m2.name)
    assert not multicodec.exists(code=m2.code)

table_cases = [
    (None, None, "identity", 0x00),
    ("multihash", None, "identity", 0x00),
    (["private", "multiaddr"], None, "ip4", 0x04),
    ("multiaddr", "permanent", "ip4", 0x04),
    (["private", "multiaddr"], ["draft", "permanent"], "ip4", 0x04),
    (None, "draft", "cidv2", 0x02),
]

@pytest.mark.parametrize("tag, status, name, code", table_cases)
def test_table(tag: Union[None, str, List[str]], status: Union[None, str, List[str]], name: str, code: int) -> None:
    """ Tests `multicodec.table`. """
    m = next(multicodec.table(tag=tag, status=status))
    assert m.name == name and m.code == code

table_failure_cases = [
    [[
        Multicodec(name="identity", tag="multihash", code=0x00, status="permanent", description="raw binary"),
        Multicodec(name="my-codec", tag="private", code=0x300001, status="draft", description="my private codec"),
    ], "Private use codes not allowed.", True],
    [[
            Multicodec(name="identity", tag="multihash", code=0x00, status="permanent", description="raw binary"),
            Multicodec(name="my-codec", tag="private", code=0x00, status="permanent", description="my private codec"),
    ], "Repeated permanent codes not allowed.", True],
    [[
            Multicodec(name="identity", tag="multihash", code=0x00, status="permanent", description="raw binary"),
            Multicodec(name="my-codec", tag="private", code=0x00, status="draft", description="my private codec"),
    ], "Repeated codes allowed if exactly one is permanent.", False],
    [[
            Multicodec(name="identity", tag="multihash", code=0x00, status="draft", description="raw binary"),
            Multicodec(name="my-codec", tag="private", code=0x00, status="permanent", description="my private codec"),
    ], "Repeated codes allowed if exactly one is permanent.", False],
    [[
            Multicodec(name="identity", tag="multihash", code=0x00, status="draft", description="raw binary"),
            Multicodec(name="my-codec", tag="private", code=0x00, status="draft", description="my private codec"),
    ], "Repeated codes allowed if exactly one is permanent.", True],
    [[
            Multicodec(name="identity", tag="multihash", code=0x00, status="permanent", description="raw binary"),
            Multicodec(name="identity", tag="private", code=0x01, status="draft", description="my private codec"),
    ], "Repeated names not allowed.", True],
]

@pytest.mark.parametrize("multicodecs, reason, fails", table_failure_cases)
def test_table_failure_modes(multicodecs: List[Multicodec], reason: str, fails: bool) -> None:
    """ Tests failure modes for codec table building. """
    try:
        multiformats_config.multicodec.build_multicodec_tables(multicodecs)
        assert not fails, reason
    except ValueError:
        assert fails, reason

# TODO: make api test parametrised

def test_api_failure_modes() -> None:
    """ Tests failure modes for the multicode API. """
    try:
        Multicodec(name="3my-codec", tag="private", code=0x00, status="draft", description="my private codec")
        assert False, "This name should not be valid."
    except ValueError:
        pass
    try:
        Multicodec(name="my-codec", tag="private", code=0x00, status="other", description="my private codec")
        assert False, "Status must be 'permanent' or 'draft'."
    except ValueError:
        pass
    try:
        Multicodec(name="my-codec", tag="private", code=-1, status="draft", description="my private codec")
        assert False, "Code must be non-negative."
    except ValueError:
        pass
    try:
        multicodec.get("my-codec")
        assert False, "Codec 'my-codec' should not exist."
    except KeyError:
        pass
    try:
        multicodec.get(code=0x300001)
        assert False, "Codec 0x300001 should not exist."
    except KeyError:
        pass
    try:
        m = Multicodec(name="identity", tag="multihash", code=0x300003, status="permanent", description="")
        multicodec.register(m)
        assert False, "Codec name 'identity' already exists."
    except ValueError:
        pass
    try:
        m = Multicodec(name="my-codec", tag="multihash", code=0x00, status="permanent", description="")
        multicodec.register(m)
        assert False, "Codec with code 0x00 already exists."
    except ValueError:
        pass



# pylint: disable = missing-docstring

import csv
import hashlib
import importlib.resources as importlib_resources
from typing import Dict, Optional

import pytest
import skein  # type: ignore
from blake3 import blake3 # type: ignore
import sha3 # type: ignore
import mmh3 # type: ignore
from Cryptodome.Hash import RIPEMD160, KangarooTwelve, SHA512

from multiformats import multihash
from multiformats.multihash import wrap, digest, unwrap

def id_digest(data: bytes, size: Optional[int]) -> bytes:
    return data if size is None else data[:size]

def sha1_digest(data: bytes, size: Optional[int]) -> bytes:
    m = hashlib.sha1()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def sha2_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m: hashlib._Hash = getattr(hashlib, f"sha{digest_bits}")()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def dbl_sha2_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m: hashlib._Hash = hashlib.sha256()
    m.update(data)
    n: hashlib._Hash = hashlib.sha256()
    n.update(m.digest())
    d = n.digest()
    return d if size is None else d[:size]

def sha3_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m: hashlib._Hash = getattr(hashlib, f"sha3_{digest_bits}")()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def sha2_512_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m = SHA512.new(truncate=str(digest_bits))
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def shake_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m: hashlib._Hash = getattr(hashlib, f'shake_{digest_bits//2}')()
    m.update(data)
    d = m.digest(digest_bits//8) # type: ignore
    return d if size is None else d[:size]

def blake2_digest(data: bytes, variant: str, digest_bits: int, size: Optional[int]) -> bytes:
    assert variant in ('b', 's')
    m: hashlib._Hash = getattr(hashlib, 'blake2{}'.format(variant))(digest_size=digest_bits//8)
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def blake3_digest(data: bytes, size: int) -> bytes:
    m = blake3()
    m.update(data)
    d: bytes = m.digest(size)
    return d

def skein_digest(data: bytes, variant: int, digest_bits: int, size: Optional[int]) -> bytes:
    assert variant in (256, 512, 1024)
    m: hashlib._Hash = getattr(skein, 'skein{}'.format(variant))(digest_bits=digest_bits)
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def keccak_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    m: hashlib._Hash = getattr(sha3, f"keccak_{digest_bits}")()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

# def murmur3_digest(data: bytes, variant: str, digest_bits: int, size: Optional[int]) -> bytes:
#     assert variant in ("32", "x64")
#     if variant == "32":
#         d: bytes = mmh3.hash(data, signed=False).to_bytes(4, byteorder="big") # pylint: disable = c-extension-no-member
#         return d if size is None else d[:size]
#     d = mmh3.hash128(data, signed=False).to_bytes(16, byteorder="big") # pylint: disable = c-extension-no-member
#     d = d[:(digest_bits//8)]
#     return d if size is None else d[:size]

def md5_digest(data: bytes, size: Optional[int]) -> bytes:
    m = hashlib.md5()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def ripemd_digest(data: bytes, digest_bits: int, size: Optional[int]) -> bytes:
    assert digest_bits == 160
    m = RIPEMD160.new()
    m.update(data)
    d = m.digest()
    return d if size is None else d[:size]

def kangarootwelve_digest(data: bytes, size: int) -> bytes:
    m = KangarooTwelve.new()
    m.update(data)
    return m.read(size)

def sha2_256_trunc254_padded_digest(data: bytes, size: Optional[int]) -> bytes:
    m: hashlib._Hash = hashlib.sha256()
    m.update(data)
    d = m.digest()
    d = d[:-1]+bytes([d[-1]&0x00111111])
    return d if size is None else d[:size]

def _test(hash_fn: str, data: bytes, hash_digest: bytes, size: Optional[int] = None) -> None:
    if size is not None:
        assert len(hash_digest) == size
    multihash_digest = digest(data, hash_fn, size=size)
    assert multihash.exists(hash_fn)
    codec = multihash.get(hash_fn)
    assert hash_fn == codec.name
    assert codec == multihash.from_digest(multihash_digest)
    assert wrap(hash_digest, hash_fn) == multihash_digest
    assert hash_digest == unwrap(multihash_digest)
    assert hash_digest == unwrap(multihash_digest, hash_fn)
    trunc_hash_digest = hash_digest[:len(hash_digest)//2]
    trunc_multihash_digest = wrap(trunc_hash_digest, hash_fn)
    assert trunc_hash_digest == unwrap(trunc_multihash_digest)
    assert trunc_hash_digest == unwrap(trunc_multihash_digest, hash_fn)
    multihash_digest = digest(bytearray(data), hash_fn, size=size)
    assert wrap(bytearray(hash_digest), hash_fn) == multihash_digest
    assert hash_digest == unwrap(bytearray(multihash_digest))
    multihash_digest = digest(memoryview(data), hash_fn, size=size)
    assert wrap(memoryview(hash_digest), hash_fn) == multihash_digest
    assert hash_digest == unwrap(memoryview(multihash_digest))

data_samples = [
    b"",
    b"Test data to be wrapped.",
    b"Test data to be wrapped."*100,
]

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 8, 16))
def test_id(data: bytes, size: Optional[int]) -> None:
    hash_fn = 'identity'
    if size is None or len(data) >= size:
        _test(hash_fn, data, id_digest(data, size), size)

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 10, 20))
def test_sha1(data: bytes, size: Optional[int]) -> None:
    hash_fn = 'sha1'
    _test(hash_fn, data, sha1_digest(data, size), size)

@pytest.mark.parametrize("digest_bits", (256, 512))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 32))
def test_sha2(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"sha2-{digest_bits}"
    _test(hash_fn, data, sha2_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("digest_bits", (256,))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 32))
def test_dbl_sha2(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"dbl-sha2-{digest_bits}"
    _test(hash_fn, data, dbl_sha2_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("digest_bits", (224, 256, 384, 512))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 28))
def test_sha3(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"sha3-{digest_bits}"
    _test(hash_fn, data, sha3_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("digest_bits", (224, 256))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 28))
def test_sha2_512(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"sha2-512-{digest_bits}"
    _test(hash_fn, data, sha2_512_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("digest_bits", (256, 512))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 32))
def test_shake(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"shake-{digest_bits//2}"
    _test(hash_fn, data, shake_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("version", ("b", "s"))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 1))
def test_blake2(data: bytes, version: str, size: Optional[int]) -> None:
    for digest_bits in range(8, (512 if version == "b" else 256)+1, 8):
        hash_fn = f"blake2{version}-{digest_bits}"
        _test(hash_fn, data, blake2_digest(data, version, digest_bits, size), size)

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (16, 32, 64))
def test_blake3(data: bytes, size: int) -> None:
    hash_fn = 'blake3'
    _test(hash_fn, data, blake3_digest(data, size), size)

@pytest.mark.parametrize("version", (256, 512, 1024))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 1))
def test_skein(data: bytes, version: int, size: Optional[int]) -> None:
    for digest_bits in range(8, version+1, 8):
        hash_fn = f"skein{version}-{digest_bits}"
        _test(hash_fn, data, skein_digest(data, version, digest_bits, size), size)

@pytest.mark.parametrize("digest_bits", (224, 256, 384, 512))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 28))
def test_keccak(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"keccak-{digest_bits}"
    _test(hash_fn, data, keccak_digest(data, digest_bits, size), size)

# @pytest.mark.parametrize("version", ("32", "x64"))
# @pytest.mark.parametrize("data", data_samples)
# @pytest.mark.parametrize("size", (None, 4))
# def test_murmur3(data: bytes, version: str, size: Optional[int]) -> None:
#     if version == "32":
#         hash_fn = f"murmur3-{version}"
#         _test(hash_fn, data, murmur3_digest(data, version, 32, size), size)
#     else:
#         for digest_bits in (64, 128):
#             hash_fn = f"murmur3-{version}-{digest_bits}"
#             _test(hash_fn, data, murmur3_digest(data, version, digest_bits, size), size)

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 8, 16))
def test_md5(data: bytes, size: Optional[int]) -> None:
    hash_fn = 'md5'
    _test(hash_fn, data, md5_digest(data, size), size)

@pytest.mark.parametrize("digest_bits", (160, ))
@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 20))
def test_ripemd(data: bytes, digest_bits: int, size: Optional[int]) -> None:
    hash_fn = f"ripemd-{digest_bits}"
    _test(hash_fn, data, ripemd_digest(data, digest_bits, size), size)

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (16, 32, 64))
def test_kangarootwelve(data: bytes, size: int) -> None:
    hash_fn = "kangarootwelve"
    _test(hash_fn, data, kangarootwelve_digest(data, size), size)

@pytest.mark.parametrize("data", data_samples)
@pytest.mark.parametrize("size", (None, 16, 32))
def test_sha_256_trunc254_padded(data: bytes, size: Optional[int]) -> None:
    hash_fn = "sha2-256-trunc254-padded"
    _test(hash_fn, data, sha2_256_trunc254_padded_digest(data, size), size)

# specific test vectors

with importlib_resources.open_text("test", "multihash-test-str-vectors.csv") as csv_table:
    multihash_test_str_vectors = list(csv.DictReader(csv_table))

@pytest.mark.parametrize("test_vector", multihash_test_str_vectors)
def test_str_vectors(test_vector: Dict[str, str]) -> None:
    hash_fn = test_vector["algorithm"]
    digest_size = int(test_vector["bits"])//8
    data = test_vector["input"].encode("utf-8")
    multihash_digest = bytes.fromhex(test_vector["multihash"])
    assert hash_fn == multihash.from_digest(multihash_digest).name
    assert digest(data, hash_fn, size=digest_size) == multihash_digest

with importlib_resources.open_text("test", "multihash-test-hex-vectors.csv") as csv_table:
    multihash_test_hex_vectors = list(csv.DictReader(csv_table))

@pytest.mark.parametrize("test_vector", multihash_test_hex_vectors)
def test_hex_vectors(test_vector: Dict[str, str]) -> None:
    hash_fn = test_vector["algorithm"]
    if hash_fn.startswith("murmur3"):
        return
    digest_size = int(test_vector["bits"])//8
    data = bytes.fromhex(test_vector["input"])
    multihash_digest = bytes.fromhex(test_vector["multihash"])
    assert hash_fn == multihash.from_digest(multihash_digest).name
    assert digest(data, hash_fn, size=digest_size) == multihash_digest



""" Tests for the `multiformats.varint` module. """


from io import BytesIO
from random import Random

import pytest
from multiformats import varint

random = Random(0)

@pytest.mark.parametrize("x", range(256))
def test_encode_decode_byte(x: int) -> None:
    """ Check that encoding followed by decoding yields the initial integer value. """
    assert varint.decode(varint.encode(x)) == x, f"Error at value {x}"


@pytest.mark.parametrize("k", range(7, 64, 7))
def test_encode_decode_bytes(k: int) -> None:
    """ Check that encoding followed by decoding yields the initial integer value. """
    for _ in range(10):
        x = 2**(k-1)+random.randrange(256)
        x_enc = varint.encode(x)
        assert varint.decode(x_enc) == x, f"Error at value {x}"
        assert varint.decode(bytearray(x_enc)) == x, f"Error at value {x} (bytearray)"


@pytest.mark.parametrize("k", range(1, 10))
def test_minimal_encoding(k: int) -> None:
    """ Check that encoding is minimal in number of bytes. """
    x = 2**(7*k)-1
    assert len(varint.encode(x)) == k, f"Error at {k} bytes"

@pytest.mark.parametrize("k", range(1, 10))
def test_stream_decoding(k: int) -> None:
    """ Check that decoding from stream only reads the varint bytes. """
    num_extra_bytes = 10
    x = 2**(7*k)-1
    stream = BytesIO(varint.encode(x)+bytes(num_extra_bytes))
    varint.decode(stream)
    assert len(stream.read()) == num_extra_bytes, f"Error at {k} bytes"

invalid_encodes = [
    (-1, "Must not encode negative integers."),
    (2**63, "Must not encode integers >= 2**63.")
]

@pytest.mark.parametrize("val, reason", invalid_encodes)
def test_encode_failure(val: int, reason: str) -> None:
    """ Checks varint encode failure modes. """
    try:
        varint.encode(val)
        assert False, reason
    except ValueError:
        pass

invalid_decodes = [
    (bytes(), "Must not decode empty byte-string."),
    (bytes([255]*9+[1]), "Must not encode varints more than 9 bytes long."),
    (bytes([255, 255]), "Must not decode invalid varint (last byte is continuation)."),
    (bytes([255, 1, 1]), "Must not leave unread bytes when decoding from bytes or bytearray objects."),
    (bytearray([255, 1, 1]), "Must not leave unread bytes when decoding from bytes or bytearray objects."),
    (bytes([0x81, 0x00]), "Must not decode non-minimally encoded varint."),
]

@pytest.mark.parametrize("val, reason", invalid_decodes)
def test_decode_failure(val: bytes, reason: str) -> None:
    """ Checks varint decode failure modes. """
    try:
        varint.decode(val)
        assert False, reason
    except ValueError:
        pass



"""
    Implementation for the ``sha`` and ``shake`` hash functions,
    using `hashlib <https://docs.python.org/3/library/hashlib.html>`_ and `pycryptodomex <https://github.com/Legrandin/pycryptodome/>`_.
"""

import hashlib
from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _sha(version: int, digest_bits: int) -> Hashfun:
    name = ("sha1", f"sha{digest_bits}", f"sha3_{digest_bits}")[version-1]
    h = getattr(hashlib, name)
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h() # pylint: disable = no-member
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_sha1(m, register) -> bool: # type: ignore
    if register is not None:
        register("sha1", _sha(1, 160), 20) # 20B = 160 bits
    return True

def _shake(digest_bits: int) -> Hashfun:
    assert digest_bits in (256, 512)
    h = getattr(hashlib, f"shake_{digest_bits//2}")
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h() # pylint: disable = no-member
        m.update(data)
        d = m.digest(digest_bits//8) # type: ignore
        return d if size is None else d[:size]
    return hashfun

def _jit_register_sha23ke(m, register) -> bool: # type: ignore
    if m[1] == "ke":
        digest_bits = 2*int(m[2])
        if register is not None:
            register(f"shake-{digest_bits//2}", _shake(digest_bits), digest_bits//8)
        return True
    sha_version, digest_bits = (int(m[1]), int(m[2]))
    if register is not None:
        register(f"sha{sha_version}-{digest_bits}", _sha(sha_version, digest_bits), digest_bits//8)
    return True

def _sha2_512(digest_bits: int) -> Hashfun:
    assert digest_bits in (224, 256)
    try:
        from Cryptodome.Hash import SHA512 # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'Cryptodome' must be installed to use the 'sha2-256' hash functions. "
                          "Consider running 'pip install pycryptodomex'.") from e
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m = SHA512.new(truncate=str(digest_bits))
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_sha2_512(m, register) -> bool: # type: ignore
    digest_bits = int(m[1])
    if register is not None:
        register(f"sha2-512-{digest_bits}", _sha2_512(digest_bits), digest_bits//8)
    return True

def _dbl_sha23(version: int, digest_bits: int) -> Hashfun:
    name = ("sha1", f"sha{digest_bits}", f"sha3_{digest_bits}")[version-1]
    h = getattr(hashlib, name)
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h() # pylint: disable = no-member
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_dbl_sha23(m, register) -> bool: # type: ignore
    sha_version, digest_bits = (int(m[1]), int(m[2]))
    if register is not None:
        register(f"sha{sha_version}-{digest_bits}", _dbl_sha23(sha_version, digest_bits), digest_bits//8)
    return True



"""
    Reports the current implementation status for multiformats.
"""
# pylint: disable = wrong-import-position, wrong-import-order, unused-import

if __name__ != "__main__":
    raise RuntimeError("usage: report.py [-h] [-d]")

# == Script imports ==

import argparse
import gc
import sys
from typing import Any, Callable, Collection, Dict, List, Optional, Tuple, Union

# `rich` is not a dependency for the `multiformats` library
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

# `setuptools_scm` is a development dependency for the `multiformats` library
from setuptools_scm import get_version # type: ignore

# `psutil` is not a dependency for the `multiformats` library
import psutil # type: ignore

# `pympler` is not a dependency for the `multiformats` library
from pympler import tracker # type: ignore

def bytesize_str(nbytes: int) -> str:
    """
        Pretty string representation of bytesizes.
    """
    sign = ""
    if nbytes < 0:
        nbytes *= -1
        sign = "-"
    suffixes = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"]
    suffix_idx = 0
    while nbytes >= 1024 and suffix_idx <= 2:
        nbytes //= 1024
        suffix_idx += 1
    return f"{sign}{nbytes}{suffixes[suffix_idx]}"

def print_diff(diff: list[tuple[str, int, int]], console: Console) -> None:
    """
        Prints a tracker diff object to console
    """
    table = Table()
    table.add_column("Type", style="bold green")
    table.add_column("Count")
    table.add_column("Size")
    for t, c, s in sorted(diff, key=lambda entry: -entry[2]):
        table.add_row(t, str(c), bytesize_str(s))
    console.print(table)

# == Extract commandline args ==

description = "Implementation report for multiformats."
parser = argparse.ArgumentParser(description=description)
parser.add_argument("-m", help='loads a minimal set of multicodecs and multibases', action="store_true")
parser.add_argument("-d", help='print codes as decimal rather than hex', action="store_true")
parser.add_argument("-r", help='saves report to file', action="store_true")
args = parser.parse_args()
minimal_load = args.m
hex_codes = not args.d
save_report = args.r
code2str: Callable[[int], str] = hex if hex_codes else str # type: ignore

# == Intro panel with version ==

version = get_version(root='.', version_scheme="post-release")

console = Console(record=True, width=110)
console.print(Panel(f"Multiformats implementation report [bold blue]v{version}[white]"))

# == Memory profiling ==

pympler_count = {}
pympler_mem_usage = {}
psutil_mem_usage = {}

gc.collect()
baseline = psutil.Process().memory_full_info().uss / (1024 * 1024)
pympler_prev = baseline
psutil_prev = baseline


tr = tracker.SummaryTracker()
tr.diff()
import typing_extensions
gc.collect()
tracker_diff = tr.diff()
pympler_count["typing-extensions"] = sum(entry[1] for entry in tracker_diff)
pypler_diff = sum(entry[2] for entry in tracker_diff)
pympler_mem_usage["typing-extensions"] = pypler_diff / (1024 * 1024)
pympler_prev += pypler_diff
psutil_diff = psutil.Process().memory_full_info().uss / (1024 * 1024)-psutil_prev
psutil_mem_usage["typing-extensions"] = psutil_diff
psutil_prev += psutil_diff

tr = tracker.SummaryTracker()
tr.diff()
import typing_validation
gc.collect()
tracker_diff = tr.diff()
pympler_count["typing-validation"] = sum(entry[1] for entry in tracker_diff)
pypler_diff = sum(entry[2] for entry in tracker_diff)
pympler_mem_usage["typing-validation"] = pypler_diff / (1024 * 1024)
pympler_prev += pypler_diff
psutil_diff = psutil.Process().memory_full_info().uss / (1024 * 1024)-psutil_prev
psutil_mem_usage["typing-validation"] = psutil_diff
psutil_prev += psutil_diff

tr = tracker.SummaryTracker()
tr.diff()
import bases
gc.collect()
tracker_diff = tr.diff()
pympler_count["bases"] = sum(entry[1] for entry in tracker_diff)
pypler_diff = sum(entry[2] for entry in tracker_diff)
pympler_mem_usage["bases"] = pypler_diff / (1024 * 1024)
pympler_prev += pypler_diff
psutil_diff = psutil.Process().memory_full_info().uss / (1024 * 1024)-psutil_prev
psutil_mem_usage["bases"] = psutil_diff
psutil_prev += psutil_diff

tr = tracker.SummaryTracker()
tr.diff()
if minimal_load:
    import multiformats_config
    multiformats_config.enable(codecs=[], bases=[])
import multiformats
from multiformats import *
gc.collect()
tracker_diff = tr.diff()
pympler_count["multiformats"] = sum(entry[1] for entry in tracker_diff)
pypler_diff = sum(entry[2] for entry in tracker_diff)
pympler_mem_usage["multiformats"] = pypler_diff / (1024 * 1024)
pympler_prev += pypler_diff
psutil_diff = psutil.Process().memory_full_info().uss / (1024 * 1024)-psutil_prev
psutil_mem_usage["multiformats"] = psutil_diff
psutil_prev += psutil_diff

pympler_mem_usage_total = sum(pympler_mem_usage.values())
pympler_mem_usage_pct = {k: v/pympler_mem_usage_total for k, v in pympler_mem_usage.items()}
psutil_mem_usage_total = sum(psutil_mem_usage.values())
psutil_mem_usage_pct = {k: v/psutil_mem_usage_total for k, v in psutil_mem_usage.items()}

# == Memory usage table ==

console.rule("Memory Usage (pympler)")

table = Table()
table.add_column("Component", style="white")
table.add_column("Obj. count", style="white", justify="right")
table.add_column("Memory", style="bold blue", justify="right")
table.add_column("Memory %", style="bold blue", justify="right")
for k, v in pympler_mem_usage.items():
    pct = f"{pympler_mem_usage_pct[k]:.0%}" if k in pympler_mem_usage_pct else ""
    if v >= 1000/1024:
        table.add_row(k, str(pympler_count[k]), f"{v:.1f}MiB", pct)
    else:
        table.add_row(k, str(pympler_count[k]), f"{1024*v:.0f}KiB", pct)
console.print(f"> memory baseline: [bold blue]{baseline:.1f}MiB[white]")
console.print(f"> multiformats memory total:     [bold blue]{pympler_mem_usage_total:.1f}MiB[white]")
console.print(table)


console.rule("Memory Usage (psutil)")

table = Table()
table.add_column("Component", style="white")
table.add_column("Memory", style="bold blue", justify="right")
table.add_column("Memory %", style="bold blue", justify="right")
for k, v in psutil_mem_usage.items():
    pct = f"{psutil_mem_usage_pct[k]:.0%}" if k in psutil_mem_usage_pct else ""
    if v >= 1000/1024:
        table.add_row(k, f"{v:.1f}MiB", pct)
    else:
        table.add_row(k, f"{1024*v:.0f}KiB", pct)
console.print(f"> memory baseline: [bold blue]{baseline:.1f}MiB[white]")
console.print(f"> multiformats memory total:     [bold blue]{psutil_mem_usage_total:.1f}MiB[white]")
console.print(table)


# == Group multihash multicodecs together ==
# TODO: consider introduce grouped multicodecs doing this directly, to reduce mem footprint (currently footprint is negligible)

_multihash_indices: Dict[str, int] = {}
_grouped_multicodecs: List[Tuple[str, str, Optional[List[int]], List[int], List[bool], typing_extensions.Literal["draft", "permanent"]]] = []
for codec in multicodec.table(tag="multihash"):
    is_implemented = multihash.is_implemented(codec.name)
    tokens = codec.name.split("-")
    if len(tokens) == 1:
        label = codec.name
    else:
        label = "-".join(tokens[:-1])
    max_digest_size: Optional[int] = None
    try:
        max_digest_size = multihash.raw.get(codec.name)[1]
    except KeyError:
        pass
    if max_digest_size is None:
        try:
            max_digest_size = int(tokens[-1])//8
        except ValueError:
            pass
    if max_digest_size is None:
        _grouped_multicodecs.append((codec.name, codec.tag, None, [codec.code], [is_implemented], codec.status))
        continue
    bitsize = max_digest_size*8
    if label not in _multihash_indices:
        _multihash_indices[label] = len(_grouped_multicodecs)
        _grouped_multicodecs.append((codec.name, codec.tag, [bitsize], [codec.code], [is_implemented], codec.status))
    else:
        bitsize_list = _grouped_multicodecs[_multihash_indices[label]][2]
        if bitsize_list is None:
            _grouped_multicodecs.append((codec.name, codec.tag, None, [codec.code], [is_implemented], codec.status))
            continue
        code_list = _grouped_multicodecs[_multihash_indices[label]][3]
        impl_list = _grouped_multicodecs[_multihash_indices[label]][4]
        bitsize_list.append(bitsize)
        code_list.append(codec.code)
        impl_list.append(is_implemented)


# == Multihash table ==

def set_str(l: Collection[int], *, use_hex: bool = False, minlen: int = 4, maxlen: int = 8) -> str:
    """ Compact representation of a sorted set of numbers. """
    assert minlen >= 3
    assert maxlen >= minlen
    tostr: Callable[[int], str] = hex if use_hex else str # type: ignore
    l = sorted(set(l))
    start = l[0]
    end = l[-1]
    step = l[1]-l[0]
    if len(l) <= minlen:
        return f"{{{', '.join(tostr(i) for i in l)}}}"
    if l == list(range(start, end+1, step)):
        fst = tostr(start)
        snd = tostr(start+step)
        lst = tostr(end)
        return f"{{{fst}, {snd}, ..., {lst}}}"
    if len(l) <= maxlen:
        return f"{{{', '.join(tostr(i) for i in l)}}}"
    return f"{{{tostr(l[0])},...(irregular)}}"


console.rule("Multihash functions")

table = Table()
table.add_column("Code", style="bold blue")
table.add_column("Name")
table.add_column("Bitsize", style="bright_black")
table.add_column("Implem.")
table.add_column("Status")
num_implemented = 0
num_total = 0
for name, tag, bitsize_list, code_list, impl_list, status in _grouped_multicodecs:
    num_total += len(impl_list)
    impl_status = "[red]no"
    if all(impl_list):
        impl_status = "[green]yes"
        num_implemented += len(impl_list)
    elif any(impl_list):
        num_impl = sum(1 if b else 0 for b in impl_list)
        impl_status = f"[yellow]{num_impl}/{len(impl_list)}"
        num_implemented += num_impl
    codec_status = "[yellow]draft" if status == "draft" else "[green]perm."
    if bitsize_list is None:
        table.add_row(code2str(code_list[0]), name, "", impl_status, codec_status)
        continue
    if len(bitsize_list) <= 1:
        table.add_row(code2str(code_list[0]), f"{name}", str(bitsize_list[0]), impl_status, codec_status)
    else:
        label = "-".join(name.split("-")[:-1])
        table.add_row(set_str(code_list, use_hex=hex_codes),
                      f"{label}-[bright_black]Bitsize",
                      set_str(bitsize_list),
                      impl_status,
                      codec_status)
console.print(f"> Multihash functions implemented: [bold blue]{num_implemented}/{num_total}")
console.print(table)


# == Multiaddr table ==

console.rule("Multiaddr protocols")
table = Table()
table.add_column("Code", style="bold blue")
table.add_column("Name")
table.add_column("Implem.")
table.add_column("Status")
num_implemented = 0
num_total = 0
for codec in multicodec.table(tag="multiaddr"):
    is_implemented = multiaddr.raw.exists(codec.name)
    num_implemented += 1 if is_implemented else 0
    num_total += 1
    impl_status = "[green]yes" if is_implemented else "[red]no"
    codec_status = "[yellow]draft" if codec.status == "draft" else "[green]perm."
    table.add_row(code2str(codec.code), codec.name, impl_status, codec_status)
console.print(f"> Multiaddr protocols implemented: [bold blue]{num_implemented}/{num_total}")
console.print(table)

# == Multibase table ==

console.rule("Multibases")
table = Table()
table.add_column("Code", style="bold blue")
table.add_column("Name")
table.add_column("Implem.")
table.add_column("Status")

num_implemented = 0
num_total = 0
for base in multibase.table():
    is_implemented = multibase.raw.exists(base.name)
    num_implemented += 1 if is_implemented else 0
    num_total += 1
    impl_status = "[green]yes" if is_implemented else "[red]no"
    codec_status = "[yellow]draft" if base.status == "draft" else "[green]perm."
    table.add_row(base.code_printable, base.name, impl_status, codec_status)
console.print(f"> Multibases implemented: [bold blue]{num_implemented}/{num_total}")
console.print(table)

# == Other multicodecs table ==

console.rule("Other Multicodecs")
table = Table()
table.add_column("Code", style="bold blue")
table.add_column("Name")
table.add_column("Tag", style="magenta")
table.add_column("Status")
for codec in multicodec.table():
    if codec.tag not in ("multihash", "multiaddr"):
        codec_status = "[yellow]draft" if codec.status == "draft" else "[green]perm."
        table.add_row(code2str(codec.code), codec.name, codec.tag, codec_status)
console.print(table)


# == Exporting report ==

if save_report:
    console.save_text("report.txt")



"""
    Implementation of the `multicodec spec <https://github.com/multiformats/multicodec>`_.

    Suggested usage:

    >>> from multiformats import multicodec
"""

import importlib.resources as importlib_resources
from io import BufferedIOBase
import json
import re
import sys
from typing import AbstractSet, Any, cast, Dict, Iterable, Iterator, Mapping, Optional, overload, Set, Sequence, Tuple, Type, TypeVar, Union
from typing_extensions import Literal
from typing_validation import validate
from multiformats_config.multicodec import load_multicodec_table

from multiformats import varint
from multiformats.varint import BytesLike
from .err import MulticodecKeyError, MulticodecValueError

def _hexcode(code: int) -> str:
    hexcode = hex(code)
    if len(hexcode) % 2 != 0:
        hexcode = "0x0"+hexcode[2:]
    return hexcode

class Multicodec:
    """
        Container class for a multicodec.

        Example usage:

        >>> Multicodec(**{
        ...     'name': 'cidv1', 'tag': 'cid', 'code': '0x01',
        ...     'status': 'permanent', 'description': 'CIDv1'})
        Multicodec(name='cidv1', tag='cid', code=1,
                   status='permanent', description='CIDv1')

        :param name: the multicodec name
        :type name: :obj:`str`
        :param tag: the multicodec tag
        :type tag: :obj:`str`
        :param code: the multicodec code, as integer or ``0xYZ`` hex-string
        :type code: :obj:`int` or :obj:`str`
        :param status: the multicodec status
        :type status: ``'draft'`` or ``'permanent'``, *optional*
        :param description: the multicodec description
        :type description: :obj:`str`, *optional*
    """

    _name: str
    _tag: str
    _code: int
    _status: Literal["draft", "permanent"]
    _description: str

    __slots__ = ("__weakref__", "_name", "_tag", "_code", "_status", "_description")

    def __new__(cls,
                name: str,
                tag: str,
                code: Union[int, str],
                status: str = "draft",
                description: str = ""
               ) -> "Multicodec":
        # pylint: disable = too-many-arguments
        for arg in (name, tag, status, description):
            validate(arg, str)
        validate(code, Union[int, str])
        name = Multicodec._validate_name(name)
        code = Multicodec.validate_code(code)
        status = Multicodec._validate_status(status)
        instance = super().__new__(cls)
        instance._name = name
        instance._tag = tag
        instance._code = code
        instance._status = status
        instance._description = description
        return instance

    @staticmethod
    def _validate_name(name: str) -> str:
        if not re.match(r"^[a-z][a-z0-9_-]+$", name):
            raise MulticodecValueError(f"Invalid multicodec name {repr(name)}")
        return name

    @staticmethod
    def validate_code(code: Union[int, str]) -> int:
        """
            Validates a multicodec code and transforms it to unsigned integer format (if in hex format).

            :param code: the multicodec code, as integer or `0xYZ` hex-string
            :type code: :obj:`int` or :obj:`str`

            :raises ValueError: if the code is invalid

        """
        if isinstance(code, str):
            if code.startswith("0x"):
                code = code[2:]
            code = int(code, base=16)
        if code < 0:
            raise MulticodecValueError(f"Invalid multicodec code {repr(code)}.")
        return code

    @staticmethod
    def _validate_status(status: str) -> Literal["draft", "permanent"]:
        if status not in ("draft", "permanent"):
            raise MulticodecValueError(f"Invalid multicodec status {repr(status)}.")
        return cast(Literal["draft", "permanent"], status)

    @property
    def name(self) -> str:
        """
            Multicodec name. Must satisfy the following:

            .. code-block:: python

                re.match(r"^[a-z][a-z0-9_-]+$", name)
        """
        return self._name

    @property
    def tag(self) -> str:
        """ Multicodec tag. """
        return self._tag

    @property
    def code(self) -> int:
        """ Multicodec code. Must be a non-negative integer. """
        return self._code

    @property
    def hexcode(self) -> str:
        """
            Multicodec code as a hex string (with hex digits zero-padded to even length):

            Example usage:

            >>> m = multicodec.get(1)
            >>> m.code
            1
            >>> m.hexcode
            '0x01'

        """
        return _hexcode(self._code)

    @property
    def status(self) -> Literal["draft", "permanent"]:
        """ Multicodec status. """
        return self._status

    @property
    def description(self) -> str:
        """ Multicodec description. """
        return self._description

    @property
    def is_private_use(self) -> bool:
        """
            Whether this multicodec code is reserved for private use,
            i.e. whether it is in ``range(0x300000, 0x400000)``.
        """
        return self.code in range(0x300000, 0x400000)

    def wrap(self, raw_data: BytesLike) -> bytes:
        """
            Wraps raw binary data into multicodec data:

            .. code-block:: console

                <raw data> --> <code><raw data>

            Example usage:

            >>> ip4 = multicodec.get("ip4")
            >>> ip4
            Multicodec(name='ip4', tag='multiaddr', code='0x04', status='permanent', description='')
            >>> raw_data = bytes([192, 168, 0, 254])
            >>> multicodec_data = ip4.wrap(raw_data)
            >>> raw_data.hex()
              'c0a800fe'
            >>> multicodec_data.hex()
            '04c0a800fe'
            >>> varint.encode(0x04).hex()
            '04' #       0x04 ^^^^ is the multicodec code for 'ip4'

            :param raw_data: the raw data to be wrapped
            :type raw_data: :obj:`~multiformats.varint.BytesLike`

            :raise ValueError: see :func:`~multiformats.varint.encode`

        """
        return varint.encode(self.code)+raw_data

    def unwrap(self, multicodec_data: BytesLike) -> bytes:
        """
            Unwraps multicodec binary data to raw data:

            .. code-block::

                <code><raw data> --> <raw data>

            Additionally checks that the code listed by the data
            matches the code of this multicodec.

            Example usage:

            >>> multicodec_data = bytes.fromhex("c0a800fe")
            >>> raw_data = ip4.unwrap(multicodec_data)
            >>> multicodec_data.hex()
            '04c0a800fe'
            >>> raw_data.hex()
              'c0a800fe'
            >>> varint.encode(0x04).hex()
            '04' #       0x04 ^^^^ is the multicodec code for 'ip4'

            :param multicodec_data: the multicodec data to be unwrapped
            :type multicodec_data: :obj:`~multiformats.varint.BytesLike`

            :raise ValueError: if the unwrapped multicodec code does not match this multicodec's code
            :raise ValueError: see :func:`multiformats.multicodec.unwrap_raw`
            :raise KeyError: see :func:`multiformats.multicodec.unwrap_raw`
        """
        code, _, raw_data = unwrap_raw(multicodec_data)
        # code, _, raw_data = varint.decode_raw(multicodec_data)
        if code != self.code:
            hexcode = _hexcode(code)
            raise MulticodecValueError(f"Found code {hexcode} when unwrapping data, expected code {self.hexcode}.")
        return bytes(raw_data)

    def to_json(self) -> Mapping[str, str]:
        """
            Returns a JSON dictionary representation of this multicodec object.

            Example usage:

            >>> m = multicodec.get(1)
            >>> m.to_json()
            {'name': 'cidv1', 'tag': 'cid', 'code': '0x01',
             'status': 'permanent', 'description': 'CIDv1'}

        """
        return {
            "name": self.name,
            "tag": self.tag,
            "code": self.hexcode,
            "status": self.status,
            "description": self.description
        }

    def __str__(self) -> str:
        if exists(self.name) and get(self.name) == self:
            return f"multicodec({repr(self.name)}, tag={repr(self.tag)})"
        return repr(self)

    def __repr__(self) -> str:
        return f"Multicodec({', '.join(f'{k}={repr(v)}' for k, v in self.to_json().items())})"

    @property
    def _as_tuple(self) -> Tuple[Type["Multicodec"], str, str, int, Literal["draft", "permanent"]]:
        return (Multicodec, self.name, self.tag, self.code, self.status)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Multicodec):
            return NotImplemented
        return self._as_tuple == other._as_tuple


def get(name: Optional[str] = None, *, code: Optional[int] = None) -> Multicodec:
    """
        Gets the multicodec with given name or code.

        Example usage:

        >>> multicodec.get("identity")
        Multicodec(name='identity', tag='multihash', code=0,
                   status='permanent', description='raw binary')
        >>> multicodec.get(code=0x01)
        Multicodec(name='cidv1', tag='ipld', code=1,
                   status='permanent', description='CIDv1')

        :param name: the multicodec name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multicodec code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises KeyError: if no such multicodec exists
        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified
    """
    validate(name, Optional[str])
    validate(code, Optional[int])
    if (name is None) == (code is None):
        raise MulticodecValueError("Must specify exactly one between 'name' and 'code'.")
    if name is not None:
        if name not in _name_table:
            raise MulticodecKeyError(f"No multicodec named {repr(name)}.")
        return _name_table[name]
    if code not in _code_table:
        raise MulticodecKeyError(f"No multicodec with code {repr(code)}.")
    return _code_table[code]


def multicodec(name: str, *, tag: Optional[str] = None) -> Multicodec:
    """
        An alias for :func:`get`, for use with multicodec name only.
        If a tag is passed, ensures that the multicodec tag matches the one given.

        Example usage:

        >>> from multiformats.multicodec import multicodec
        >>> multicodec("identity")
        Multicodec(name='identity', tag='multihash', code=0,
                   status='permanent', description='raw binary')

        :param name: the multicodec name
        :type name: :obj:`str`
        :param tag: the optional multicodec tag
        :type tag: :obj:`str` or :obj:`None`, *optional*

        :raises KeyError: see :func:`get`
    """
    codec = get(name)
    if tag is not None and codec.tag != tag:
        raise MulticodecKeyError(f"Multicodec {repr(name)} exists, but its tag is not {repr(tag)}.")
    return codec


def exists(name: Union[None, str, Multicodec] = None, *, code: Optional[int] = None) -> bool:
    """
        Checks whether there is a multicodec with the given name or code.

        Example usage:

        >>> multicodec.exists("identity")
        True
        >>> multicodec.exists(code=0x01)
        True

        :param name: the multicodec name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multicodec code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified
    """
    validate(name, Optional[str])
    validate(code, Optional[int])
    if (name is None) == (code is None):
        raise MulticodecValueError("Must specify exactly one between 'name' and 'code'.")
    if name is not None:
        return name in _name_table
    return code in _code_table


def wrap(codec: Union[str, int, Multicodec], raw_data: BytesLike) -> bytes:
    """
        Wraps raw binary data into multicodec data:

        .. code-block::

            <raw data> --> <code><raw data>

        Example usage:

        >>> raw_data = bytes([192, 168, 0, 254])
        >>> multicodec_data = multicodec.wrap("ip4", raw_data)
        >>> raw_data.hex()
          'c0a800fe'
        >>> multicodec_data.hex()
        '04c0a800fe'
        >>> varint.encode(0x04).hex()
        '04' #       0x04 ^^^^ is the multicodec code for 'ip4'

        :param codec: the multicodec that the raw data refers to
        :type codec: :obj:`str`, :obj:`int` or :class:`Multicodec`
        :param raw_data: the raw binary data
        :type raw_data: :obj:`~multiformats.varint.BytesLike`

        :raises KeyError: see :func:`get`
    """
    if isinstance(codec, str):
        codec = get(codec)
    elif isinstance(codec, int):
        codec = get(code=codec)
    else:
        validate(codec, Multicodec)
    return codec.wrap(raw_data)

def unwrap(multicodec_data: BytesLike) -> Tuple[Multicodec, bytes]:
    """
        Unwraps multicodec binary data to multicodec and raw data:

        Example usage:

        >>> multicodec_data = bytes.fromhex("04c0a800fe")
        >>> codec, raw_data = multicodec.unwrap(multicodec_data)
        >>> multicodec_data.hex()
        '04c0a800fe'
        >>> raw_data.hex()
          'c0a800fe'
        >>> codec
        Multicodec(name='ip4', tag='multiaddr', code='0x04', status='permanent', description='')

        :param multicodec_data: the binary data prefixed with multicodec code
        :type multicodec_data: :obj:`~multiformats.varint.BytesLike`

        :raises KeyError: if the code does not correspond to a know multicodec
    """
    code, _, raw_data = unwrap_raw(multicodec_data)
    return get(code=code), bytes(raw_data)


_BufferedIOT = TypeVar("_BufferedIOT", bound=BufferedIOBase)

@overload
def unwrap_raw(multicodec_data: BytesLike) -> Tuple[int, int, memoryview]:
    ...

@overload
def unwrap_raw(multicodec_data: _BufferedIOT) -> Tuple[int, int, _BufferedIOT]:
    ...

def unwrap_raw(multicodec_data: Union[BytesLike, BufferedIOBase]) -> Tuple[int, int, Union[memoryview, BufferedIOBase]]:
    """
        Similar to :func:`unwrap`, but returns a triple of multicodec code, number of bytes read and remaining bytes.

        Example usage:

        >>> multicodec_data = bytes.fromhex("04c0a800fe")
        >>> code, num_bytes_read, remaining_bytes = multicodec.unwrap_raw(multicodec_data)
        >>> code
        4
        >>> num_bytes_read
        1
        >>> remaining_bytes
        <memory at 0x000001BE46B17640>
        >>> multicodec_data.hex()
        '04c0a800fe'
        >>> bytes(remaining_bytes).hex()
          'c0a800fe'

        :param multicodec_data: the binary data prefixed with multicodec code
        :type multicodec_data: :obj:`~multiformats.varint.BytesLike`

        :raises KeyError: if the code does not correspond to a know multicodec
    """
    code, n, raw_data = varint.decode_raw(multicodec_data)
    if not exists(code=code):
        raise MulticodecKeyError(f"No multicodec is known with unwrapped code {_hexcode(code)}.")
    return code, n, raw_data


def validate_multicodec(codec: Multicodec) -> None:
    """
        Validates an instance of :class:`Multicodec`.
        If the multicodec is registered (i.e. valid), no error is raised.

        :param codec: the instance to be validated
        :type codec: :class:`Multicodec`

        :raises KeyError: if no multicodec with the given name is registered
        :raises ValueError: if a multicodec with the given name is registered, but is different from the one given

    """
    validate(codec, Multicodec)
    mc = get(codec.name)
    if mc != codec:
        raise MulticodecValueError(f"Multicodec named {codec.name} exists, but is not the one given.")

def register(codec: Multicodec, *, overwrite: bool = False) -> None:
    """
        Registers a given multicodec.

        Example usage:

        >>> m = Multicodec("my-multicodec", "my-tag", 0x300001, "draft", "...")
        >>> multicodec.register(m)
        >>> multicodec.exists(code=0x300001)
        True
        >>> multicodec.get(code=0x300001).name
        'my-multicodec'
        >>> multicodec.get(code=0x300001).is_private_use
        True

        :param codec: the multicodec to register
        :type codec: :class:`Multicodec`
        :param overwrite: whether to overwrite a multicodec with existing code (optional, default :obj:`False`)
        :type overwrite: :obj:`bool`, *optional*

        :raises ValueError: if ``overwrite`` is :obj:`False` and a multicodec with the same name or code already exists
        :raises ValueError: if ``overwrite`` is :obj:`True` and a multicodec with the same name but different code already exists
    """
    validate(codec, Multicodec)
    validate(overwrite, bool)
    if not overwrite and codec.code in _code_table:
        raise MulticodecValueError(f"Multicodec with code {repr(codec.code)} already exists: {_code_table[codec.code]}")
    if codec.name in _name_table and _name_table[codec.name].code != codec.code:
        raise MulticodecValueError(f"Multicodec with name {repr(codec.name)} already exists: {_name_table[codec.name]}")
    _code_table[codec.code] = codec
    _name_table[codec.name] = codec


def unregister(name: Optional[str] = None, *, code: Optional[int] = None) -> None:
    """
        Unregisters the multicodec with given name or code.

        Example usage:

        >>> multicodec.unregister(code=0x01) # cidv1
        >>> multicodec.unregister(code=0x01)
        False

        :param name: the multicodec name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multicodec code
        :type code: :obj:`int` or :obj:`None`, *optional*

        :raises KeyError: if no such multicodec exists
        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified
    """
    m = get(name, code=code)
    del _code_table[m.code]
    del _name_table[m.name]



def table(*,
          tag: Union[None, str, AbstractSet[str], Sequence[str]] = None,
          status: Union[None, str, AbstractSet[str], Sequence[str]] = None) -> Iterator[Multicodec]:
    """
        Iterates through the registered multicodecs, in order of ascending code.

        Example usage:

        >>> len(list(multicodec.table())) # multicodec.table() returns an iterator
        482
        >>> selected = multicodec.table(tag=["cid", "cid", "multiaddr"], status="permanent")
        >>> [m.code for m in selected]
        [1, 4, 6, 41, 53, 54, 55, 56, 81, 85, 112, 113, 114, 120,
         144, 145, 146, 147, 148, 149, 150, 151, 152, 176, 177,
         178, 192, 193, 290, 297, 400, 421, 460, 477, 478, 479, 512]

        :param tag: one or more tags to be selected (if :obj:`None`, all tags are included)
        :type tag: :obj:`None`, :obj:`str`, set or sequence of :obj:`str`, *optional*
        :param status: one or more statuses to be selected (if :obj:`None`, all statuses are included)
        :type status: :obj:`None`, :obj:`str`, set or sequence of :obj:`str`, *optional*

    """
    validate(tag, Union[None, str, AbstractSet[str], Sequence[str]])
    validate(status, Union[None, str, AbstractSet[str], Sequence[str]])
    tags: Union[None, AbstractSet[str], Sequence[str]]
    if tag is None:
        tags = None
    elif isinstance(tag, str):
        tags = [tag]
    else:
        tags = tag
    statuses: Union[None, AbstractSet[str], Sequence[str]]
    if status is None:
        statuses = None
    elif isinstance(status, str):
        statuses = [status]
    else:
        statuses = status
    for code in sorted(_code_table.keys()):
        m = _code_table[code]
        if tags is not None and m.tag not in tags:
            continue
        if statuses is not None and m.status not in statuses:
            continue
        yield m

_code_table, _name_table = load_multicodec_table()

# def build_multicodec_tables(codecs: Iterable[Multicodec], *,
#                             allow_private_use: bool = False) -> Tuple[Dict[int, Multicodec], Dict[str, Multicodec]]:
#     """
#         Creates code->multicodec and name->multicodec mappings from a finite iterable of multicodecs,
#         returning the mappings.

#         Example usage:

#         >>> code_table, name_table = build_multicodec_tables(codecs)

#         :param codecs: multicodecs to be registered
#         :type codecs: iterable of :class:`Multicodec`
#         :param allow_private_use: whether to allow multicodec entries with private use codes in ``range(0x300000, 0x400000)`` (default :obj:`False`)
#         :type allow_private_use: :obj:`bool`, *optional*

#         :raises ValueError: if ``allow_private_use`` and a multicodec with private use code is encountered
#         :raises ValueError: if the same multicodec code is encountered multiple times, unless exactly one of the multicodecs
#         has permanent status (in which case that codec is the one inserted in the table)
#         :raises ValueError: if the same name is encountered multiple times

#     """
#     # validate(codecs, Iterable[Multicodec]) # TODO: not yet properly supported by typing-validation
#     validate(allow_private_use, bool)
#     code_table: Dict[int, Multicodec] = {}
#     name_table: Dict[str, Multicodec] = {}
#     overwritten_draft_codes: Set[int] = set()
#     for m in codecs:
#         if not allow_private_use and m.is_private_use:
#             raise MulticodecValueError(f"Private use multicodec not allowed: {m}")
#         if m.code in code_table:
#             if code_table[m.code].status == "permanent":
#                 if m.status == "draft":
#                     # this draft code has been superseded by a permanent one, skip it
#                     continue
#                 raise MulticodecValueError(f"Multicodec code {m.hexcode} appears multiple times in table.")
#             if m.status != "permanent":
#                 # overwriting draft code with another draft code: dodgy, need to check at the end
#                 overwritten_draft_codes.add(m.code)
#         code_table[m.code] = m
#         if m.name in name_table:
#             raise MulticodecValueError(f"Multicodec name {m.name} appears multiple times in table.")
#         name_table[m.name] = m
#     for code in overwritten_draft_codes:
#         m = code_table[code]
#         if m.status != "permanent":
#             raise MulticodecValueError(f"Code {m.code} appears multiple times in table, "
#                               "but none of the associated multicodecs is permanent.")
#     return code_table, name_table

# Create the global code->multicodec and name->multicodec mappings.
# _code_table: Dict[int, Multicodec] = {}
# _name_table: Dict[str, Multicodec] = {}
# with importlib_resources.open_text("multiformats.multicodec", "multicodec-table.json") as _table_f:
#     _table_json = json.load(_table_f)
#     _code_table, _name_table = build_multicodec_tables(Multicodec(**row) for row in _table_json)



"""
    Implementation of the `multibase spec <https://github.com/multiformats/multibase>`_.

    Suggested usage:

    >>> from multiformats import multibase
"""

from abc import ABC, abstractmethod
import binascii
import importlib.resources as importlib_resources
from itertools import product
import json
import math
import re
from typing import Any, Callable, cast, Dict, Iterable, Iterator, List, Mapping, Optional, Tuple, Type, Union
import sys

from typing_extensions import Literal
from typing_validation import validate

from bases import (base2, base16, base8, base10, base36, base58btc, base58flickr, base58ripple,
                   base32, base32hex, base32z, base64, base64url, base45,)
from multiformats_config.multibase import load_multibase_table

from multiformats.multibase import raw
from multiformats.varint import BytesLike
from .raw import RawEncoder, RawDecoder
from .err import MultibaseKeyError, MultibaseValueError

class Multibase:
    """
        Container class for a multibase encoding.

        Example usage:

        >>> Multibase(name="base16", code="f",
                      status="default", description="hexadecimal")

        :param name: the multibase name
        :type name: :obj:`str`
        :param code: the multibase code, as single-char string or ``0x...`` hex-string of a non-empty bytestring
        :type code: :obj:`str`
        :param status: the multibase status
        :type status: ``'draft'``, ``'candidate'`` or ``'default'``, *optional*
        :param description: the multibase description
        :type description: :obj:`str`, *optional*
    """

    _name: str
    _code: str
    _status: Literal["draft", "candidate", "default"]
    _description: str

    __slots__ = ("__weakref__", "_name", "_code", "_status", "_description")

    def __new__(cls,
                name: str,
                code: str,
                status: str = "draft",
                description: str = ""
                ) -> "Multibase":
        for arg in (name, code, status, description):
            validate(arg, str)
        name = Multibase._validate_name(name)
        code = Multibase.validate_code(code)
        status = Multibase._validate_status(status)
        instance = super().__new__(cls)
        instance._name = name
        instance._code = code
        instance._status = status
        instance._description = description
        return instance

    @staticmethod
    def _validate_name(name: Optional[str]) -> str:
        validate(name, Optional[str])
        assert name is not None
        if not re.match(r"^[a-z][a-z0-9_-]+$", name): # ensures len(name) > 1
            raise MultibaseValueError(f"Invalid multibase encoding name {repr(name)}")
        return name

    @staticmethod
    def validate_code(code: str) -> str:
        r"""
            Validates a multibase code and transforms it to single-character format (if in hex format).

            Example usage:

            >>> Multibase.validate_code("0x00")
            '\x00'
            >>> Multibase.validate_code("hi")
            MultibaseValueError: Multibase codes must be single-character strings
            or the hex digits '0xYZ' of a single byte.

            :param code: the multibase code, as single character or ``0x...`` hex-string of a non-empty bytestring
            :type code: :obj:`str`

            :raises ValueError: if the code is invalid

        """
        validate(code, str)
        if re.match(r"^0x([0-9a-zA-Z][0-9a-zA-Z])+$", code):
            ord_code = int(code, base=16)
            if ord_code in range(0x20, 0x7F):
                raise MultibaseValueError("Multibase codes in hex format cannot be printable ASCII characters.")
            code = chr(ord_code)
        elif len(code) != 1:
            raise MultibaseValueError("Multibase codes must be single-character strings or the hex digits '0x...' of a non-empty bytestring.")
        return code

    @staticmethod
    def _validate_status(status: str) -> Literal["draft", "candidate", "default"]:
        if status not in ("draft", "candidate", "default"):
            raise MultibaseValueError(f"Invalid multibase encoding status {repr(status)}.")
        return cast(Literal["draft", "candidate", "default"], status)

    @property
    def code(self) -> str:
        """
            Multibase code. Must either have length 1 or satisfy:

            .. code-block:: python

               re.match(r"^0x$", code)

        """
        return self._code

    @property
    def code_printable(self) -> str:
        r"""
            Printable version of :meth:`Multibase.code`:

            - if the code is a single non-printable ASCII character, returns the hex string of its byte
            - otherwise, returns the code itself

            Example usage:

            >>> identity = multibase.get(code="\x00")
            >>> identity.code
            '\x00'
            >>> identity.code_printable
            '0x00'

        """
        code = self.code
        ord_code = ord(code)
        if ord_code not in range(0x20, 0x7F):
            ord_code_num_bytes = max(1, math.ceil(ord_code.bit_length()/8))
            ord_code_bytes = ord_code.to_bytes(ord_code_num_bytes, byteorder="big")
            return "0x"+base16.encode(ord_code_bytes)
        return code

    @property
    def status(self) -> Literal["draft", "candidate", "default"]:
        """ Multibase status. Must be ``'draft'``, ``'candidate'`` or ``'default'``."""
        return self._status

    @property
    def description(self) -> str:
        """ Multibase description. """
        return self._description

    @property
    def name(self) -> str:
        """
            Multibase name. Must satisfy the following:

            .. code-block:: python

                re.match(r"^[a-z][a-z0-9_-]+$", name)

            In the `multibase table <https://github.com/multiformats/multibase/raw/master/multibase.csv>`_,
            this is listed under `encoding`.
        """
        return self._name

    @property
    def raw_encoder(self) -> RawEncoder:
        """
            Returns the raw encoder for this encoding: given bytes, it produces the encoded string without the multibase prefix.
        """
        enc = raw.get(self.name)
        if enc is None:
            raise NotImplementedError(f"Multibase/decoding for {repr(self.name)} is not yet implemented.")
        return enc.encode

    @property
    def raw_decoder(self) -> RawDecoder:
        """
            Returns the raw encoder for this encoding: given a string without the multibase prefix, it produces the decoded data.
        """
        enc = raw.get(self.name)
        if enc is None:
            raise NotImplementedError(f"Multibase/decoding for {repr(self.name)} is not yet implemented.")
        return enc.decode

    def encode(self, b: BytesLike) -> str:
        """
            Encodes bytes into a multibase string: it first uses :meth:`Multibase.raw_encoder`,
            and then prepends the multibase prefix given by :attr:`Multibase.code` and returns the resulting multibase string.

            Example usage:

            >>> base32 = multibase.get("base32")
            >>> base32.encode(b"Hello World!")
            'bjbswy3dpeblw64tmmqqq'

            :param b: the bytes to be encoded
            :type s: :class:`~multiformats.varint.BytesLike`

        """
        return self.code+self.raw_encoder(b)

    def decode(self, s: str) -> bytes:
        """
            Decodes a multibase string into bytes: it first checks that the multibase
            prefix matches the value specified by :attr:`Multibase.code`, then uses
            :meth:`Multibase.raw_decoder` on the string without prefix and returns the bytes.

            Example usage:

            >>> base32 = multibase.get("base32")
            >>> base32.decode("bjbswy3dpeblw64tmmqqq")
            b'Hello World!'

            :param s: the string to be decoded
            :type s: :obj:`str`

            :raises ValueError: if the code from the string is different from the one of this multibase
            :raises ValueError: see :func:`from_str`
            :raises KeyError: see :func:`from_str`
        """
        encoding = from_str(s)
        if encoding != self:
            raise MultibaseValueError(f"Expected {repr(self.name)} encoding, "
                                 f"found {repr(encoding.name)} encoding instead.")
        return self.raw_decoder(s[1:])

    def to_json(self) -> Mapping[str, str]:
        """
            Returns a JSON dictionary representation of this :class:`Multibase` object.

            Example usage:

            >>> base32 = multibase.get("base32")
            >>> base32.to_json()
            {'name': 'base32', 'code': 'b',
             'status': 'default',
             'description': 'rfc4648 case-insensitive - no padding'}

        """
        return {
            "name": self.name,
            "code": self.code_printable,
            "status": self.status,
            "description": self.description
        }

    def __str__(self) -> str:
        if exists(self.name) and get(self.name) == self:
            return f"multibase.get({repr(self.name)})"
        return repr(self)

    def __repr__(self) -> str:
        return f"Multibase({', '.join(f'{k}={repr(v)}' for k, v in self.to_json().items())})"

    @property
    def _as_tuple(self) -> Tuple[Type["Multibase"], str, str, Literal["draft", "candidate", "default"]]:
        return (Multibase, self.name, self.code, self.status)

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, Multibase):
            return NotImplemented
        return self._as_tuple == other._as_tuple


def get(name: Optional[str] = None, *, code: Optional[str] = None) -> Multibase:
    """
        Gets the multibase encoding with given name or multibase code.

        Example usage:

        >>> multibase.get("base8")
        Multibase(encoding='base8', code='7',
                  status='draft', description='octal')
        >>> multibase.get(name="base8")
        Multibase(encoding='base8', code='7',
                  status='draft', description='octal')
        >>> multibase.get(code="t")
        Multibase(encoding='base32hexpad', code='t', status='candidate',
                  description='rfc4648 case-insensitive - with padding')

        :param name: the name of this multibase
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the code of this multibase (keyword-only)
        :type name: :obj:`str` or :obj:`None`, *optional*

        :raises ValueError: if the empty string is passed
        :raises KeyError: if no such multibase exists
        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified

    """
    validate(name, Optional[str])
    validate(code, Optional[str])
    if (name is None) == (code is None):
        raise MultibaseValueError("Must specify exactly one between encoding name and code.")
    if code is not None:
        if code not in _code_table:
            raise MultibaseKeyError(f"No multibase encoding with code {repr(code)}.")
        return _code_table[code]
    if name not in _name_table:
        raise MultibaseKeyError(f"No multibase encoding named {repr(name)}.")
    return _name_table[name]


def exists(name: Optional[str] = None, *, code: Optional[str] = None) -> bool:
    """
        Checks whether a multibase encoding with given name or code exists.

        Example usage:

        >>> multibase.exists("base8")
        True
        >>> multibase.exists(code="t")
        True

        :param name: the name of this multibase
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the code of this multibase (keyword-only)
        :type name: :obj:`str` or :obj:`None`, *optional*

        :raises ValueError: if the empty string is passed
        :raises ValueError: unless exactly one of ``name`` and ``code`` is specified
    """
    validate(name, Optional[str])
    validate(code, Optional[str])
    if (name is None) == (code is None):
        raise MultibaseValueError("Must specify exactly one between encoding name and code.")
    if code is not None:
        code = Multibase.validate_code(code)
        return code in _code_table
    return name in _name_table


def register(base: Multibase, *, overwrite: bool = False) -> None:
    """
        Registers a given multibase encoding.

        Example usage:

        >>> base45 = Multibase(name="base45", code=":",
                               status="draft", description="base45 encoding")
        >>> multibase.register(base45)
        >>> multibase.get("base45")
        Multibase(encoding='base45', code=':', status='draft',
                  description='base45 encoding')

        :param base: the multibase to register
        :type base: :class:`Multibase`
        :param overwrite: whether to overwrite a multibase with existing code (optional, default :obj:`False`)
        :type overwrite: :obj:`bool`, *optional*

        :raises ValueError: if ``overwrite`` is :obj:`False` and a multibase with the same name or code already exists
        :raises ValueError: if ``overwrite`` is :obj:`True` and a multibase with the same name but different code already exists

    """
    validate(base, Multibase)
    validate(overwrite, bool)
    if not overwrite and base.code in _code_table:
        raise MultibaseValueError(f"Multibase encoding with code {repr(base.code)} already exists: {_code_table[base.code]}")
    if base.name in _name_table and _name_table[base.name].code != base.code:
        raise MultibaseValueError(f"Multibase encoding with name {repr(base.name)} already exists: {_name_table[base.name]}")
    _code_table[base.code] = base
    _name_table[base.name] = base


def validate_multibase(multibase: Multibase) -> None:
    """
        Validates an instance of :class:`Multibase`.
        If the multibase is registered (i.e. valid), no error is raised.

        :param multibase: the instance to be validated
        :type multibase: :class:`Multibase`

        :raises KeyError: if no multibase with the given name is registered
        :raises ValueError: if a multibase with the given name is registered, but is different from the one given
    """
    validate(multibase, Multibase)
    mc = get(multibase.name)
    if mc != multibase:
        raise MultibaseValueError(f"Multibase named {multibase.name} exists, but is not the one given.")


def unregister(name: Optional[str] = None, *, code: Optional[str] = None) -> None:
    """
        Unregisters the multibase encoding with given name or code.

        Example usage:

        >>> base45 = Multibase(name="base45", code=":",
                               status="draft", description="base45 encoding")
        >>> multibase.register(base45)
        >>> multibase.get("base45")
        Multibase(encoding='base45', code=':', status='draft',
                  description='base45 encoding')
        >>> multibase.unregister(code=":")
        >>> multibase.exists("base45")
        False

        :param name: the multibase name
        :type name: :obj:`str` or :obj:`None`, *optional*
        :param code: the multibase code
        :type code: :obj:`str` or :obj:`None`, *optional*

        :raises KeyError: if no such multibase exists
    """
    enc = get(name=name, code=code)
    del _code_table[enc.code]
    del _name_table[enc.name]


def table() -> Iterator[Multibase]:
    """
        Iterates through the registered multibases, in order of ascending code.

        Example usage:

        >>> [e.code for e in multibase.table()]
        ['\\x00', '0', '7', '9', 'B', 'C', 'F', 'K', 'M', 'T', 'U', 'V',
         'Z','b', 'c', 'f', 'h', 'k', 'm', 'p', 't', 'u', 'v', 'z']

    """
    for code in sorted(_code_table.keys()):
        yield _code_table[code]


def from_str(s: str) -> Multibase:
    """
        Returns the multibase encoding for the given string, according to the code specified by its prefix.

        Example usage:

        >>> multibase.from_str("mSGVsbG8gd29ybGQh")
        Multibase(encoding='base64', code='m', status='default',
                  description='rfc4648 no padding')

        :param s: the multibase encoded string
        :type s: :obj:`str`

        :raises ValueError: if the empty string is passed
        :raises KeyError: if no multibase exists with that code
    """
    validate(s, str)
    if len(s) == 0:
        raise MultibaseValueError("Empty string is not valid for encoded data.")
    if s[0] in _code_table:
        return _code_table[s[0]]
    for code in _code_table:
        if s.startswith(code):
            return get(code=code)
    raise MultibaseKeyError("No known multibase code is a prefix of the given string.")


def encode(data: BytesLike, base: Union[str, "Multibase"]) -> str:
    """
        Encodes the given bytes into a multibase string using the given encoding.

        If the encoding is passed by name, the :func:`get` function is used to retrieve it.
        Multibase encoding is performed by the :meth:`multiformats.multibase.Multibase.encode` method.

        Example usage:

        >>> multibase.encode(b"Hello world!", "base64")
        'mSGVsbG8gd29ybGQh'

        :param data: the data to encode using the multibase
        :type data: :obj:`~multiformats.varint.BytesLike`
        :param base: the multibase to use
        :type base: :obj:`str` or :class:`Multibase`
    """
    validate(base, Union[str, "Multibase"])
    if isinstance(base, str):
        base = get(base)
    return base.encode(data)


def decode(s: str) -> bytes:
    """
        Decodes the given multibase string into bytes.
        The encoding is inferred using the :func:`from_str` function.
        Decoding is then performed by :meth:`Multibase.decode` method.

        Example usage:

        >>> multibase.decode("mSGVsbG8gd29ybGQh")
        b'Hello world!'

        :param s: the string to be decoded
        :type s: :obj:`str`
    """
    base = from_str(s)
    return base.decode(s)


def decode_raw(s: str) -> Tuple[Multibase, bytes]:
    """
        Similar to :func:`decode`, but returns a ``(base, bytestr)`` pair
        of the multibase and decoded bytestring.

        Example usage:

        >>> base, bytestr = multibase.decode_raw("mSGVsbG8gd29ybGQh")
        >>> base
        Multibase(name='base64', code='m',
                  status='default', description='rfc4648 no padding')
        >>> bytestr
        b'Hello world!'

        :param s: the string to be decoded
        :type s: :obj:`str`

    """
    base = from_str(s)
    return base, base.decode(s)

_code_table, _name_table = load_multibase_table()

# def build_multibase_tables(bases: Iterable[Multibase]) -> Tuple[Dict[str, Multibase], Dict[str, Multibase]]:
#     """
#         Creates code->encoding and name->encoding mappings from a finite iterable of encodings, returning the mappings.

#         Example usage:

#         >>> code_table, name_table = build_multicodec_tables(bases)

#         :param bases: the multibases to add to the table
#         ::

#         :raises ValueError: if the same encoding code or name is encountered multiple times
#     """
#     # validate(multicodecs, Iterable[Multicodec]) # TODO: not yet properly supported by typing-validation
#     code_table: Dict[str, Multibase] = {}
#     name_table: Dict[str, Multibase] = {}
#     for e in bases:
#         if e.code in code_table:
#             raise MultibaseValueError(f"Multicodec name {e.name} appears multiple times in table.")
#         code_table[e.code] = e
#         if e.name in name_table:
#             raise MultibaseValueError(f"Multicodec name {e.name} appears multiple times in table.")
#         name_table[e.name] = e
#     return code_table, name_table

# Create the global code->multibase and name->multibase mappings.
# _code_table: Dict[str, Multibase] = {}
# _name_table: Dict[str, Multibase] = {}
# with importlib_resources.open_text("multiformats.multibase", "multibase-table.json", encoding="utf8") as _table_f:
#     _table_json = json.load(_table_f)
#     _code_table, _name_table = build_multibase_tables(Multibase(**row) for row in _table_json)



"""
    Implementation for the ``keccak`` hash functions, using the optional dependency `pysha3 <https://github.com/tiran/pysha3>`_.
"""

import hashlib
from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _keccak(digest_bits: int) -> Hashfun:
    try:
        import sha3 # type: ignore # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'sha3' must be installed to use 'keccak' hash functions. Consider running 'pip install pysha3'.") from e
    h = getattr(sha3, f"keccak_{digest_bits}")
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, digest_bits//8)
        m: hashlib._Hash = h() # pylint: disable = no-member
        m.update(data)
        d = m.digest()
        return d if size is None else d[:size]
    return hashfun

def _jit_register_keccak(m, register) -> bool: # type: ignore
    digest_bits = int(m[1])
    if register is not None:
        register(f"keccak-{digest_bits}", _keccak(digest_bits), digest_bits//8)
    return True



"""
    Implementation of raw encodings used by multiaddr protocols.

    For expected address bytestring sizes, see the `multiaddr table <https://github.com/multiformats/multiaddr/blob/master/protocols.csv>`_.
"""

from ipaddress import IPv4Address, IPv6Address, AddressValueError
from typing import Callable, Dict, Optional, Tuple
from typing_validation import validate

from multiformats.varint import BytesLike
from .err import MultiaddrKeyError, MultiaddrValueError

RawEncoder = Callable[[str], bytes]
""" Type alias for raw address value encoders. """

RawDecoder = Callable[[BytesLike], str]
""" Type alias for raw address value decoders. """

ProtoImpl = Tuple[Optional[RawEncoder], Optional[RawDecoder], Optional[int]]
""" Type alias for raw protocol implementation, as a triple ``(raw_encoder, raw_decoder, addr_size)``. """

_proto_impl: Dict[str, ProtoImpl] = {}

def get(name: str) -> ProtoImpl:
    """
        Gets the implementation ``(raw_encoder, raw_decoder, addr_size)`` for a protocol with given name.

        The ``addr_size`` component is the size in bytes for the binary representation of protocol addresses:

        - for protocols with no address, ``addr_size`` is 0
        - for protocols with addresses of variable binary size, ``addr_size`` is :obj:`None`
        - for all other protocols, ``addr_size`` is a positive :obj:`int`

        Example usage:

        >>> multiaddr.raw.get("ip4")
        (
         <function ip4_encoder at 0x000002DDE1655550>,
         <function ip4_decoder at 0x000002DDE16555E0>,
         4
        )

        :param name: the protocol implementation name
        :type name: :obj:`str`

        :raises KeyError: if no such protocol implementation exists
    """
    validate(name, str)
    if name not in _proto_impl:
        raise MultiaddrKeyError(f"No implementation for protocol {repr(name)}.")
    return _proto_impl[name]


def exists(name: str) -> bool:
    """
        Checks whether the protocol with given name has an implementation.

        Example usage:

        >>> multiaddr.raw.exists("ip4")
        True

        :param name: the protocol implementation name
        :type name: :obj:`str`
    """
    validate(name, str)
    return name in _proto_impl


def register(name: str, raw_encoder: Optional[RawEncoder], raw_decoder: Optional[RawDecoder], addr_size: Optional[int], *,
             overwrite: bool = False) -> None:
    """
        Registers an implementation for the protocol by given name.

        If ``addr_size`` is 0, ``raw_encoder`` and ``raw_decoder`` should both be :obj:`None` (because the protocol admits no address).

        It is expected that ``raw_encoder`` raises :class:`~multiformats.multiaddr.err.MultiaddrValueError`
        if the string passed to it is not a valid string representatio for an address of this protocol.
        It is expected that ``raw_decoder`` raises :class:`~multiformats.multiaddr.err.MultiaddrValueError`
        if the bytestring passed to it is not a valid binary representation for an address of this protocol.

        Example usage for protocol requiring address value:

        .. code-block:: python

            def ip4_encoder(s: str) -> bytes:
                validate(s, str)
                return IPv4Address(s).packed

            def ip4_decoder(b: BytesLike) -> str:
                validate(b, BytesLike)
                _validate_size('ip4', b, 4)
                return str(IPv4Address(b))

            multiformats.raw.register("ip4", ip4_encoder, ip4_decoder, 4)


        Example usage for protocol not requiring address value:

        .. code-block:: python

            multiformats.raw.register("quic", None, None, 0)

        :param name: the protocol implementation name
        :type name: :obj:`str`
        :param raw_encoder: the raw encoder
        :type raw_encoder: :obj:`RawEncoder` or :obj:`None`
        :param raw_decoder: the raw decoder
        :type raw_decoder: :obj:`RawDecoder` or :obj:`None`
        :param addr_size: the expected address size for protocol addresses, in bytes (0 if no address is expected, :obj:`None` if address size is variable)
        :type addr_size: :obj:`int` or :obj:`None`
        :param overwrite: whether to overwrite an existing implementation with the same name
        :type overwrite: :obj:`bool`

        :raises ValueError: if ``addr_size`` is a negative integer
        :raises ValueError: if ``addr_size`` is 0 and either one of ``raw_encoder`` or ``raw_decoder`` is not :obj:`None`
        :raises ValueError: if ``overwrite`` is :obj:`False` and an implementation with the same name already exists
    """
    validate(name, str)
    # validate(raw_encoder, Optional[RawEncoder]) # TODO: typing-validation does not yet support this
    # validate(raw_decoder, Optional[RawDecoder]) # TODO: typing-validation does not yet support this
    validate(addr_size, Optional[int])
    validate(overwrite, bool)
    if addr_size is not None and addr_size < 0:
        raise MultiaddrValueError("Size must be None or non-negative integer.")
    if addr_size == 0 and (raw_encoder is not None or raw_decoder is not None):
        raise MultiaddrValueError("Protocol admits no address (addr_size=0), set raw encoder and decoder to None.")
    if not overwrite and name in _proto_impl:
        raise MultiaddrValueError(f"Implementation for protocol {repr(name)} already exists.")
    _proto_impl[name] = (raw_encoder, raw_decoder, addr_size)


def unregister(name: str) -> None:
    """
        Unregisters the implementatio for the protocol by given name.

        Example usage:

        >>> multiformats.raw.unregister("ip4")
        >>> multiformats.raw.exists("ip4")
        False

        :param name: the protocol implementation name
        :type name: :obj:`str`

        :raises KeyError: if no such protocol implementation exists
    """
    validate(name, str)
    if name not in _proto_impl:
        raise MultiaddrKeyError(f"Implementation for protocol {repr(name)} does not exist.")
    del _proto_impl[name]

def _validate_size(name: str, b: BytesLike, size: int) -> None:
    if len(b) != size:
        raise MultiaddrValueError(f"Incorrect length for {repr(name)} bytes: found {len(b)}, expected {size}.")

def ip4_encoder(s: str) -> bytes:
    """ Encoder for 'ip4' protocol. """
    validate(s, str)
    try:
        return IPv4Address(s).packed
    except AddressValueError as e:
        raise MultiaddrValueError(str(e)) from e

def ip4_decoder(b: BytesLike) -> str:
    """ Decoder for 'ip4' protocol. """
    validate(b, BytesLike)
    _validate_size('ip4', b, 4)
    try:
        return str(IPv4Address(b))
    except AddressValueError as e:
        raise MultiaddrValueError(str(e)) from e

register("ip4", ip4_encoder, ip4_decoder, 4)

def ip6_encoder(s: str) -> bytes:
    """ Encoder for 'ip6' protocol. """
    validate(s, str)
    try:
        return IPv6Address(s).packed
    except AddressValueError as e:
        raise MultiaddrValueError(str(e)) from e

def ip6_decoder(b: BytesLike) -> str:
    """ Decoder for 'ip6' protocol. """
    validate(b, BytesLike)
    _validate_size('ip6', b, 16)
    try:
        return str(IPv6Address(b))
    except AddressValueError as e:
        raise MultiaddrValueError(str(e)) from e

register("ip6", ip6_encoder, ip6_decoder, 16)

def tcp_udp_encoder(s: str) -> bytes:
    """ Encoder for 'tcp' and 'udp' protocols. """
    validate(s, str)
    if not s.isdigit():
        raise MultiaddrValueError(f"Invalid UDP port {repr(s)}.")
    x = int(s)
    if not 0 <= x < 65536:
        raise MultiaddrValueError(f"UDP port {repr(s)} out of range.")
    return x.to_bytes(2, byteorder="big")

def tcp_udp_decoder(b: BytesLike) -> str:
    """ Decoder for 'tcp' and 'udp' protocol. """
    validate(b, BytesLike)
    _validate_size('udp', b, 2)
    return str(int.from_bytes(b, byteorder="big"))

register("tcp", tcp_udp_encoder, tcp_udp_decoder, 2)
register("udp", tcp_udp_encoder, tcp_udp_decoder, 2)

# TODO: dccp, 2 bytes
# TODO: ip6zone, variable, rfc4007 IPv6 zone
# TODO: dns, variable
# TODO: dns4, variable
# TODO: dns6, variable
# TODO: dnsaddr, variable
# TODO: sctp, 2 bytes
# TODO: unix, variable
# TODO: p2p, variable
# TODO: ipfs, variable, backwards compatibility; equivalent to /p2p
# TODO: onion, 12 bytes
# TODO: onion3, 37 bytes
# TODO: garlic64, variable
# TODO: garlic32, variable
# TODO: memory, variable, in memory transport for self-dialing and testing; arbitrary

# Protocols without address value:
register("udt", None, None, 0)
register("utp", None, None, 0)
register("tls", None, None, 0)
register("noise", None, None, 0)
register("quic", None, None, 0)
register("http", None, None, 0)
register("https", None, None, 0) # deprecated alias for /tls/http
register("ws", None, None, 0) # WebSockets
register("wss", None, None, 0) # deprecated alias for /tls/ws
register("p2p-websocket-star", None, None, 0)
register("p2p-stardust", None, None, 0)
register("p2p-webrtc-star", None, None, 0)
register("p2p-webrtc-direct", None, None, 0)
register("p2p-circuit", None, None, 0)



"""  Tests for the `multiformats.multibase` module. """
# pylint: disable = missing-docstring

# import pytest
# TODO: make tests parametrised

from random import Random

import multiformats_config
from multiformats import multibase
from multiformats.multibase import Multibase


def test_exists() -> None:
    """ Tests `multibase.exists`. """
    assert multibase.exists("identity")
    assert multibase.exists(code="\x00")

def test_get() -> None:
    """ Tests `multibase.get`. """
    enc = multibase.get("base16")
    assert enc.name == "base16"
    assert enc.code == "f"
    assert enc.status == "default"
    assert enc.description == "hexadecimal"

def test_multibase_contructor() -> None:
    """ Tests `Multibase.from_json`. """
    m_json = {
        "name": "my-codec",
        "code": "W",
        "status": "draft",
        "description": "my private codec"
    }
    enc = Multibase(**m_json)
    assert enc.name == m_json["name"]
    assert enc.code == "W"
    assert enc.status == m_json["status"]
    assert enc.description == m_json["description"]
    assert str(enc) == f"Multibase({', '.join(f'{k}={repr(v)}' for k, v in m_json.items())})"
    m_json = {
        "name": "my-codec",
        "code": "0x02",
        "status": "draft",
        "description": "my private codec"
    }
    enc = Multibase(**m_json)
    assert enc.name == m_json["name"]
    assert enc.code == "\x02"
    assert enc.status == m_json["status"]
    assert enc.description == m_json["description"]
    assert str(enc) == f"Multibase({', '.join(f'{k}={repr(v)}' for k, v in m_json.items())})"


def test_multibase_to_json() -> None:
    """ Tests `Multibase.to_json`. """
    m_json = {
        "name": "my-codec",
        "code": "W",
        "status": "draft",
        "description": "my private codec"
    }
    assert Multibase(**m_json).to_json() == m_json
    m_json = {
        "name": "my-codec",
        "code": "0x02",
        "status": "draft",
        "description": "my private codec"
    }
    assert Multibase(**m_json).to_json() == m_json

def test_register() -> None:
    """ Tests `multibase.register`. """
    m_json = {
        "name": "my-codec",
        "code": "W",
        "status": "draft",
        "description": "my private codec"
    }
    enc = Multibase(**m_json)
    assert not multibase.exists(enc.name)
    assert not multibase.exists(code=enc.code)
    multibase.register(enc)
    assert multibase.exists(enc.name)
    assert multibase.exists(code=enc.code)
    assert multibase.get(enc.name) == multibase.get(code=enc.code) == enc

def test_unregister() -> None:
    """ Tests `multibase.get`. """
    m_json = {
        "name": "my-codec",
        "code": "W",
        "status": "draft",
        "description": "my private codec"
    }
    enc = Multibase(**m_json)
    assert multibase.exists(enc.name)
    assert multibase.exists(code=enc.code)
    m2_json = {
        "name": "my-codec-2",
        "code": "0x02",
        "status": "draft",
        "description": "my second private codec"
    }
    m2 = Multibase(**m2_json)
    assert not multibase.exists(m2.name)
    assert not multibase.exists(code=m2.code)
    multibase.register(m2)
    multibase.unregister(enc.name)
    assert not multibase.exists(enc.name)
    assert not multibase.exists(code=enc.code)
    multibase.unregister(code=m2.code)
    assert not multibase.exists(m2.name)
    assert not multibase.exists(code=m2.code)

def test_table() -> None:
    """ Tests `multibase.table`. """
    try:
        enc = next(multibase.table())
        assert enc.name == "identity" and enc.code_printable == "0x00"
    except StopIteration:
        assert False, "At least one multibase exists."
    try:
        multibases = [
            Multibase(name="identity", code="0x00", status="default", description=""),
            Multibase(name="my-codec", code="0x00", status="draft", description=""),
        ]
        multiformats_config.multibase.build_multibase_tables(multibases)
        assert False, "Repeated codes not allowed."
    except ValueError:
        pass
    try:
        multibases = [
            Multibase(name="identity", code="0x00", status="default", description=""),
            Multibase(name="identity", code="0x01", status="draft", description=""),
        ]
        print(multiformats_config.multibase.build_multibase_tables(multibases))
        assert False, "Repeated names not allowed."
    except ValueError:
        pass

def test_api_failure_modes() -> None:
    """ Tests failure modes for the multicode API. """
    # pylint: disable = too-many-statements
    try:
        Multibase(name="3my-codec", code="0x00")
        assert False, "This name should not be valid."
    except ValueError:
        pass
    try:
        Multibase(name="my-codec", code="0x00", status="other")
        assert False, "Status must be 'draft', 'candidate' or 'default'."
    except ValueError:
        pass
    try:
        Multibase(name="my-codec", code="0x79")
        assert False, "Codes in hex format must be non-printable ASCII characters."
    except ValueError:
        pass
    try:
        Multibase(name="3my-codec", code="0x20")
        assert False, "Codes in hex format must be non-printable ASCII characters."
    except ValueError:
        pass
    try:
        Multibase(name="3my-codec", code="\x00")
        assert False, "Codes in single-character format must be printable ASCII characters."
    except ValueError:
        pass
    try:
        Multibase(name="3my-codec", code="\x80")
        assert False, "Codes in single-character format must be printable ASCII characters."
    except ValueError:
        pass
    try:
        Multibase(name="3my-codec", code="\x7F")
        assert False, "Codes in single-character format must be printable ASCII characters."
    except ValueError:
        pass
    try:
        multibase.get("my-codec")
        assert False, "Multibase 'my-codec' should not exist."
    except KeyError:
        pass
    try:
        multibase.get(code='\x08')
        assert False, "Multibase with code 'W' should not exist."
    except KeyError:
        pass
    try:
        enc = Multibase(name="identity", code="0x01", status="default")
        multibase.register(enc)
        assert False, "Multibase name 'identity' already exists."
    except ValueError:
        pass
    try:
        enc = Multibase(name="my-codec", code="\x00", status="draft")
        multibase.register(enc)
        assert False, "Codec with code 0x00 already exists."
    except ValueError:
        pass

rand = Random(0)
nsamples = 1024

_proquints = {
    bytes([127, 0, 0, 1]): "lusab-babad",
    bytes([63, 84, 220, 193]): "gutih-tugad",
    bytes([63, 118, 7, 35]): "gutuk-bisog",
    bytes([140, 98, 193, 141]): "mudof-sakat",
    bytes([64, 255, 6, 200]): "haguz-biram",
    bytes([128, 30, 52, 45]): "mabiv-gibot",
    bytes([147, 67, 119, 2]): "natag-lisaf",
    bytes([212, 58, 253, 68]): "tibup-zujah",
    bytes([216, 35, 68, 215]): "tobog-higil",
    bytes([216, 68, 232, 21]): "todah-vobij",
    bytes([198, 81, 129, 136]): "sinid-makam",
    bytes([12, 110, 110, 204]): "budov-kuras",
}
for b, s in list(_proquints.items()):
    _proquints[b[:2]] = s[:5]
    _proquints[b[2:]] = s[6:]

def test_proquint() -> None:
    proquint = multibase.get("proquint")
    for b, s in _proquints.items():
        s = 'pro-'+s
        error_msg = f"Proquint encode error at b = {list(b)}, s = {repr(s)}"
        assert proquint.encode(b) == s
        error_msg = f"Proquint decode error at b = {list(b)}, s = {repr(s)}"
        assert proquint.decode(s) == b
    assert proquint.encode(b"") == "pro-"
    assert proquint.decode("pro-") == b""
    for idx in range(nsamples):
        l = rand.randint(1, 5)
        i = rand.randrange(0, 256**l)
        b = i.to_bytes(l, byteorder="big")
        s = proquint.encode(b)
        error_msg = f"Proquint decode-encode error at sample #{idx}: b = {list(b)}, s = {repr(s)}"
        assert proquint.decode(s) == b, error_msg



"""
    Implementation for the ``kangarootwelve`` hash function, using `pycryptodomex <https://github.com/Legrandin/pycryptodome/>`_.
"""

from typing import Optional

from multiformats.varint import BytesLike
from .utils import Hashfun, validate_hashfun_args

def _kangarootwelve() -> Hashfun:
    try:
        from Cryptodome.Hash import KangarooTwelve # pylint: disable = import-outside-toplevel
    except ImportError as e:
        raise ImportError("Module 'Cryptodome' must be installed to use the 'kangarootwelve' hash function. "
                          "Consider running 'pip install pycryptodomex'.") from e
    def hashfun(data: BytesLike, size: Optional[int] = None) -> bytes:
        validate_hashfun_args(data, size, None, size_required=True, name="kangarootwelve")
        assert size is not None
        m = KangarooTwelve.new()
        m.update(data)
        return m.read(size)
    return hashfun

def _jit_register_kangarootwelve(m, register) -> bool: # type: ignore
    if register is not None:
        register("kangarootwelve", _kangarootwelve(), None)
    return True



"""
    Errors for the :mod:`~multiformats.multicodec` module.
"""

import builtins

class MulticodecKeyError(builtins.KeyError): # pylint: disable = redefined-builtin
    """ Class for :mod:`~multiformats.multicodec` key errors. """

class MulticodecValueError(builtins.ValueError): # pylint: disable = redefined-builtin
    """ Class for :mod:`~multiformats.multicodec` value errors. """



"""
    Implementation for the ``sha2-256-trunc254-padded`` hash function,
    using `hashlib <https://docs.python.org/3/library/hashlib.html>`_.

    Future support planned for the ``poseidon-bls12_381-a2-fc1`` hash functions,
    possibly using `poseidon-hash <https://github.com/ingonyama-zk/poseidon-hash>`_.
    Additional references on the Poseidon hash function:

    - https://www.poseidon-hash.info/
    - https://github.com/filecoin-project/neptune
"""

import hashlib
from hashlib import sha256
from typing import Optional

from multiformats.varint import BytesLike
from .utils import validate_hashfun_args

def _sha_256_trunc254_padded(data: BytesLike, size: Optional[int] = None) -> bytes:
    validate_hashfun_args(data, size, 32)
    m: hashlib._Hash = sha256() # pylint: disable = no-member
    m.update(data)
    d = m.digest()
    d = d[:-1]+bytes([d[-1]&0x00111111])
    return d if size is None else d[:size]

def _jit_register_sha_256_trunc254_padded(m, register) -> bool: # type: ignore
    if register is not None:
        register("sha2-256-trunc254-padded", _sha_256_trunc254_padded, 32) # 32B = 256 bits
    return True



# pylint: disable = missing-docstring

from random import Random
import sys
from typing import List, Union

import pytest

from multiformats import multicodec, multiaddr
from multiformats.multicodec import Multicodec
from multiformats.multicodec.err import MulticodecKeyError
from multiformats.multiaddr import Proto, Addr, Multiaddr
from multiformats.multiaddr.err import MultiaddrKeyError, MultiaddrValueError


implemented_protocols: List[Proto] = []

@pytest.mark.parametrize("multicodec", multicodec.table(tag="multiaddr"))
def test_implemented_protos(multicodec: Multicodec) -> None:
    name = multicodec.name
    if multiaddr.raw.exists(name):
        implemented_protocols.append(Proto(name))

invalid_multiaddr_strings = [
    "/ip4",
    "/ip4/::1",
    "/ip4/fdpsofodsajfdoisa",
    "/ip6",
    "/ip6zone",
    "/ip6zone/",
    "/ip6zone//ip6/fe80::1",
    "/udp",
    "/tcp",
    "/sctp",
    "/udp/65536",
    "/tcp/65536",
    "/onion/9imaq4ygg2iegci7:80",
    "/onion/aaimaq4ygg2iegci7:80",
    "/onion/timaq4ygg2iegci7:0",
    "/onion/timaq4ygg2iegci7:-1",
    "/onion/timaq4ygg2iegci7",
    "/onion/timaq4ygg2iegci@:666",
    "/onion3/9ww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:80",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd7:80",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:0",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:a",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:-1",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyy@:666",
    "/udp/1234/sctp",
    "/udp/1234/udt/1234",
    "/udp/1234/utp/1234",
    "/ip4/127.0.0.1/udp/jfodsajfidosajfoidsa",
    "/ip4/127.0.0.1/udp",
    "/ip4/127.0.0.1/tcp/jfodsajfidosajfoidsa",
    "/ip4/127.0.0.1/tcp",
    "/ip4/127.0.0.1/p2p",
    "/ip4/127.0.0.1/p2p/tcp",
    "/unix",
    "/ip4/1.2.3.4/tcp/80/unix",
    "/ip4/127.0.0.1/tcp/9090/http/p2p-webcrt-direct",
    "/dns",
    "/dns4",
    "/dns6"
]

@pytest.mark.parametrize("multiaddr_str", invalid_multiaddr_strings)
def test_invalid_parse(multiaddr_str: str) -> None:
    try:
        multiaddr.parse(multiaddr_str)
        assert False
    except MultiaddrValueError: # invalid string
        pass
    except MultiaddrKeyError: # protocol not implemented
        pass
    except MulticodecKeyError: # protocol does not exist
        pass

valid_multiaddr_strings = [
    "/ip4/1.2.3.4",
    "/ip4/0.0.0.0",
    "/ip6/::1",
    "/ip6/2601:9:4f81:9700:803e:ca65:66e8:c21",
    "/ip6zone/x/ip6/fe80::1",
    "/ip6zone/x%y/ip6/fe80::1",
    "/ip6zone/x%y/ip6/::",
    "/onion/timaq4ygg2iegci7:1234",
    "/onion/timaq4ygg2iegci7:80/http",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:1234",
    "/onion3/vww6ybal4bd7szmgncyruucpgfkqahzddi37ktceo3ah7ngmcopnpyyd:80/http",
    "/udp/0",
    "/tcp/0",
    "/sctp/0",
    "/udp/1234",
    "/tcp/1234",
    "/sctp/1234",
    "/utp",
    "/udp/65535",
    "/tcp/65535",
    "/p2p/QmcgpsyWgH8Y8ajJz1Cu72KnS5uo2Aa2LpzU7kinSupNKC",
    "/udp/1234/sctp/1234",
    "/udp/1234/udt",
    "/udp/1234/utp",
    "/tcp/1234/http",
    "/tcp/1234/https",
    "/p2p/QmcgpsyWgH8Y8ajJz1Cu72KnS5uo2Aa2LpzU7kinSupNKC/tcp/1234",
    "/ip4/127.0.0.1/udp/1234",
    "/ip4/127.0.0.1/p2p/QmcgpsyWgH8Y8ajJz1Cu72KnS5uo2Aa2LpzU7kinSupNKC/tcp/1234",
    "/unix/a/b/c/d/e",
    "/unix/Überrschung!/大柱",
    "/unix/stdio",
    "/ip4/1.2.3.4/tcp/80/unix/a/b/c/d/e/f",
    "/ip4/127.0.0.1/p2p/QmcgpsyWgH8Y8ajJz1Cu72KnS5uo2Aa2LpzU7kinSupNKC"
    "/tcp/1234/unix/stdio",
    "/ip4/127.0.0.1/tcp/9090/http/p2p-webrtc-direct",
    "/dns/example.com",
    "/dns4/موقع.وزارة-الاتصالات.مصر"
]

@pytest.mark.parametrize("multiaddr_str", valid_multiaddr_strings)
def test_valid_parse(multiaddr_str: str) -> None:
    try:
        multiaddr.parse(multiaddr_str)
    except multiaddr.err.MultiaddrKeyError: # protocol not implemented
        pass

@pytest.mark.parametrize("multiaddr_str", valid_multiaddr_strings)
def test_parse_str(multiaddr_str: str) -> None:
    try:
        ma = multiaddr.parse(multiaddr_str)
        str(ma)
        assert str(ma) == multiaddr_str
    except multiaddr.err.MultiaddrKeyError: # protocol not implemented
        pass

@pytest.mark.parametrize("multiaddr_str", valid_multiaddr_strings)
def test_bytes_decode(multiaddr_str: str) -> None:
    try:
        ma = multiaddr.parse(multiaddr_str)
        bytes(ma)
        assert ma == multiaddr.decode(bytes(ma))
    except multiaddr.err.MultiaddrKeyError: # protocol not implemented
        pass

rand = Random(0)
num_attempts = 10

def random_addr(proto: Proto) -> Union[Proto, Addr]:
    addr_size = proto.addr_size
    if addr_size == 0:
        return proto
    for _ in range(num_attempts):
        if addr_size is None:
            addr_size = rand.randint(1, 8)
        if sys.version_info[1] >= 9:
            addr_bytes = rand.randbytes(addr_size)
        else:
            addr_bytes = rand.getrandbits(addr_size*8).to_bytes(addr_size, byteorder="big")
        if not proto.is_addr_valid(addr_bytes):
            with pytest.raises(multiaddr.err.MultiaddrValueError):
                proto/addr_bytes # pylint: disable = pointless-statement
            continue
        return proto/addr_bytes
    raise RuntimeError(f"Could not generate a valid random address for protocol {repr(proto.name)}")

def random_multiaddr(use_slash: bool = False, maxlen: int = 4) -> Multiaddr:
    assert maxlen >= 1
    l = rand.randint(1, maxlen)
    protos = rand.sample(implemented_protocols, l)
    addrs = []
    for idx in range(l):
        addrs.append(random_addr(protos[idx]))
    if not use_slash:
        return Multiaddr(*addrs)
    a = Multiaddr(addrs[0])
    for idx in range(1, l):
        a = a/addrs[idx]
    return a

@pytest.mark.parametrize("sample", range(40))
@pytest.mark.parametrize("use_slash", [False, True])
def test_random(sample: int, use_slash: bool) -> None:
    ma = random_multiaddr(use_slash=use_slash)
    str(ma)
    assert ma == multiaddr.parse(str(ma))
    bytes(ma)
    assert ma == multiaddr.decode(bytes(ma))
